from collections import defaultdict
from .base import Permission, Role, Scope
from .checker import has_permission


class PermissionRegistry:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._permissions: dict[str, Permission] = {}
            cls._instance._roles: dict[str, Role] = {}
            cls._instance._scopes: dict[str, Scope] = {}
        return cls._instance

    def register_scope(
        self,
        key: str,
        label: str,
        permissions: list[str],
        description: str = '',
        is_oidc: bool = False,
    ):
        if key not in self._scopes:
            scope = Scope(key, label, set(permissions), description, is_oidc)
            self._scopes[key] = scope

    def register_role(self, code: str, name: str, description: str = ''):
        """Define a system role structure."""
        if code not in self._roles:
            role = Role(code, name=name, description=description, permissions=set())
            self._roles[code] = role

    def register_permission(
        self,
        key: str,
        label: str,
        module: str,
        description: str = '',
        default_roles: list[str] = None,
    ):
        """
        Register a permission and optionally assign it to default roles.
        """
        if key in self._permissions:
            raise ValueError(f'Permission {key} already registered')

        # 1. Store the permission metadata
        perm = Permission(key, label=label, description=description, module=module)
        self._permissions[key] = perm

        # 2. Map to roles if provided
        if default_roles:
            for role_code in default_roles:
                self.assign_to_role(role_code, key)

    def assign_to_role(self, role_code: str, permission: str):
        """Manually assign a pattern to a role."""
        if role_code not in self._roles:
            self.register_role(role_code, role_code.capitalize())

        role = self._roles[role_code]
        role.permissions.add(permission)

    def assign_to_scopes(self, scope_key: str, permissions: list[str]):
        if scope_key not in self._scopes:
            raise ValueError(f'Scope {scope_key} not registered')
        self._scopes[scope_key].permissions.update(permissions)

    def get_permission_keys(self):
        return self._permissions.keys()

    def get_permission_list(self):
        return self._permissions.values()

    def get_role(self, code: str):
        return self._roles.get(code)

    def get_role_codes(self):
        return list(self._roles.keys())

    def get_role_list(self):
        return self._roles.values()

    def get_scope_keys(self):
        return self._scopes.keys()

    def get_scope_list(self):
        return self._scopes.values()

    def get_grouped_permissions(self):
        grouped = defaultdict(list)
        for perm in self._permissions.values():
            grouped[perm.module].append(perm)
        return dict(grouped)

    def get_permissions_for_scopes(self, scopes):
        perms: set[str] = set()
        for key in scopes:
            if key in self._scopes:
                obj = self._scopes[key]
                perms.update(obj.permissions)
        return perms

    def get_scopes_for_permission(self, perm_key: str) -> list[str]:
        matching_scopes = []
        for scope_key, scope_obj in self._scopes.items():
            # If the scope's patterns match the required permission
            if has_permission(perm_key, scope_obj.permissions):
                matching_scopes.append(scope_key)
        return matching_scopes


perm_registry = PermissionRegistry()



import requests

# Fetch disposable email domains from
# https://github.com/disposable-email-domains/disposable-email-domains

blocklist_url = 'https://raw.githubusercontent.com/disposable-email-domains/disposable-email-domains/refs/heads/main/disposable_email_blocklist.conf'


def fetch_blocklist():
    resp = requests.get(blocklist_url)
    with open('src/saas_base/rules/blocklist.txt', 'w') as f:
        f.write(resp.text)


if __name__ == '__main__':
    fetch_blocklist()



from django.urls import path
from ..endpoints.invitations import (
    InvitationListEndpoint,
    InvitationItemEndpoint,
)

urlpatterns = [
    path('', InvitationListEndpoint.as_view()),
    path('<pk>/', InvitationItemEndpoint.as_view()),
]



from dataclasses import dataclass


@dataclass
class Permission:
    key: str
    label: str
    description: str
    module: str


@dataclass
class Role:
    code: str
    name: str
    description: str
    permissions: set[str]


@dataclass
class Scope:
    key: str
    label: str
    permissions: set[str]  # Internal permission patterns
    description: str = ''
    is_oidc: bool = False  # Flag for OIDC standard scopes



from django.db import models
from saas_base.db.fields import EncryptedField


class UserSecret(models.Model):
    secret_key = EncryptedField(null=True, blank=True)


class TenantSecret(models.Model):
    tenant = models.ForeignKey('saas_base.Tenant', on_delete=models.CASCADE)
    secret_key = EncryptedField(null=True, blank=True)



from rest_framework.request import Request
from rest_framework.mixins import (
    ListModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
)
from saas_base.drf.views import TenantEndpoint
from saas_base.drf.decorators import resource_permission
from saas_base.drf.filters import TenantIdFilter, IncludeFilter, ChoiceFilter
from saas_base.serializers.member import (
    MemberSerializer,
    MemberUpdateSerializer,
)
from saas_base.models import Member

__all__ = [
    'MemberListEndpoint',
    'MemberItemEndpoint',
]


class MemberListEndpoint(ListModelMixin, TenantEndpoint):
    serializer_class = MemberSerializer
    filter_backends = [TenantIdFilter, IncludeFilter]
    queryset = Member.objects.all()

    include_select_related_fields = ['user']
    include_prefetch_related_fields = ['groups']

    @resource_permission('iam.member.view')
    def get(self, request: Request, *args, **kwargs):
        """List all members in the tenant."""
        return self.list(request, *args, **kwargs)


class MemberItemEndpoint(UpdateModelMixin, DestroyModelMixin, TenantEndpoint):
    serializer_class = MemberUpdateSerializer
    queryset = Member.objects.all()

    @resource_permission('iam.member.manage')
    def patch(self, request: Request, *args, **kwargs):
        """Update a member's permissions and groups."""
        return self.partial_update(request, *args, **kwargs)

    @resource_permission('iam.member.manage')
    def delete(self, request: Request, *args, **kwargs):
        """Remove a member from the tenant."""
        # TODO: prevent removing owner
        return self.destroy(request, *args, **kwargs)



from rest_framework.settings import api_settings
from saas_base.settings import saas_settings

__all__ = ['get_client_ip']


CLIENT_IP_HEADERS = (
    'Forwarded-For',  # Standard header defined by RFC 7239.
    'X-Forwarded-For',  # Common header for proxies, load balancers, like AWS ELB. Default to the left-most IP.
    'CF-Connecting-IP',  # Used by CloudFlare.
    'True-Client-IP',  # Header for CloudFlare Enterprise.
    'X-Real-IP',  # Header used by some providers like Amazon EC2, Heroku.
    'Client-IP',  # Header used by some providers like Amazon EC2, Heroku.
    'X-Client-IP',  # Used by Microsoft Azure.
    'Forwarded',  # Standard header defined by RFC 7239.
    'X-Forwarded',  # Used by Squid and similar software.
    'X-Cluster-Client-IP',  # Used by Rackspace Cloud Load Balancers.
    'Fastly-Client-IP',  # Used by Fastly, Firebase.
)


def get_client_ip(request, ip_headers=None):
    remote_addr = request.META.get('REMOTE_ADDR')

    if ip_headers is None:
        if saas_settings.CLIENT_IP_HEADERS:
            ip_headers = saas_settings.CLIENT_IP_HEADERS
        else:
            ip_headers = CLIENT_IP_HEADERS

    for key in ip_headers:
        ip = request.headers.get(key)
        if ip:
            if ',' in ip:
                return _parse_proxied_ip(ip, remote_addr)
            return ip

    return remote_addr


def _parse_proxied_ip(xff: str, remote_addr: str):
    num_proxies = api_settings.NUM_PROXIES

    if num_proxies is not None:
        if num_proxies == 0 or xff is None:
            return remote_addr

        addresses = xff.split(',')
        client_addr = addresses[-min(num_proxies, len(addresses))]
        return client_addr.strip()

    return ''.join(xff.split()) if xff else remote_addr



from django.core.management.base import BaseCommand
from saas_base.models import get_tenant_model, Member


class Command(BaseCommand):
    help = 'Ensure all tenant owners have Member records'

    def handle(self, *args, **options):
        self.stdout.write('Checking tenant owner integrity...')
        tenants = get_tenant_model().objects.all()
        created_count = 0
        updated_count = 0

        for tenant in tenants:
            # Check if the owner has a member record in their own tenant
            member, created = Member.objects.get_or_create(tenant=tenant, user=tenant.owner, defaults={'role': 'OWNER'})

            if created:
                created_count += 1
            elif member.role != 'OWNER':
                # Force owner to have OWNER role if they have a different one
                member.role = 'OWNER'
                member.save()
                updated_count += 1

        self.stdout.write(
            self.style.SUCCESS(
                f'Integrity check complete: Created {created_count} missing owner members, '
                f'Updated {updated_count} owner roles.'
            )
        )



from tests.client import FixturesTestCase


class TestCurrentTenantAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_fetch_demo_tenant(self):
        self.force_login()
        resp = self.client.get('/m/tenant/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['slug'], self.tenant.slug)



from django.db import transaction
from django.utils import timezone
from rest_framework.exceptions import PermissionDenied
from rest_framework.request import Request
from rest_framework.response import Response

from saas_base.drf.views import AuthenticatedEndpoint
from saas_base.drf.decorators import resource_permission
from saas_base.models import Invitation, Member
from saas_base.serializers.invitation import InvitationInfoSerializer

__all__ = [
    'UserInvitationListEndpoint',
    'UserInvitationAcceptEndpoint',
]


class UserInvitationListEndpoint(AuthenticatedEndpoint):
    serializer_class = InvitationInfoSerializer
    queryset = Invitation.objects.select_related('tenant', 'inviter').all()
    pagination_class = None

    def get_queryset(self):
        queryset = self.queryset.filter(email=self.request.user.email)
        return queryset.filter(status=Invitation.InviteStatus.SENT)

    @resource_permission('user.org.view')
    def get(self, request: Request, *args, **kwargs):
        """List all the current user's invitations."""
        queryset = self.filter_queryset(self.get_queryset())

        items = []
        with transaction.atomic():
            for invitation in queryset:
                if invitation.is_expired():
                    invitation.status = Invitation.InviteStatus.EXPIRED
                    invitation.save()
                else:
                    items.append(invitation)

        serializer = self.get_serializer(items, many=True)
        return Response(serializer.data)


class UserInvitationAcceptEndpoint(AuthenticatedEndpoint):
    serializer_class = InvitationInfoSerializer
    queryset = Invitation.objects.all()

    def join_tenant(self, invitation: Invitation):
        if self.request.user.email != invitation.email:
            raise PermissionDenied(' This invitation is not for you.')

        if Member.objects.filter(user=self.request.user, tenant_id=invitation.tenant_id).count():
            invitation.status = Invitation.InviteStatus.ACCEPTED
            invitation.accepted_at = timezone.now()
            invitation.save()
            return

        if invitation.is_expired():
            invitation.status = Invitation.InviteStatus.EXPIRED
            invitation.save()
            return

        with transaction.atomic():
            member = Member.objects.create(
                user=self.request.user,
                tenant_id=invitation.tenant_id,
                role=invitation.role,
            )
            member.groups.set(invitation.groups.all())
            invitation.status = Invitation.InviteStatus.ACCEPTED
            invitation.accepted_at = timezone.now()
            invitation.save()

    @resource_permission('user.org.join')
    def patch(self, request: Request, *args, **kwargs):
        status = request.data.get('status')
        obj = self.get_object()
        if status == 'accepted':
            self.join_tenant(obj)

        serializer = self.get_serializer(obj)
        return Response(serializer.data)



from saas_base.models import Invitation
from tests.client import FixturesTestCase


class TestInvitationsAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def setUp(self):
        super().setUp()
        self.invitation = Invitation.objects.create(
            tenant=self.tenant,
            inviter_id=self.OWNER_USER_ID,
            email='invitee@example.com',
            role='MEMBER',
            status=Invitation.InviteStatus.SENT,
        )

    def test_list_invitations(self):
        self.force_login()
        url = f'/m/invitations/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], str(self.invitation.id))

    def test_create_invitation(self):
        self.force_login()
        url = f'/m/invitations/'
        data = {
            'email': 'new@example.com',
            'role': 'MEMBER',
        }
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, 201)

        invitation = Invitation.objects.get(email='new@example.com')
        self.assertEqual(invitation.tenant_id, self.tenant_id)
        self.assertEqual(invitation.role, 'MEMBER')
        self.assertEqual(invitation.inviter_id, self.user_id)

    def test_update_invitation(self):
        self.force_login()
        url = f'/m/invitations/{self.invitation.id}/'
        data = {
            'role': 'ADMIN',
        }
        resp = self.client.patch(url, data)
        self.assertEqual(resp.status_code, 200)

        self.invitation.refresh_from_db()
        self.assertEqual(self.invitation.role, 'ADMIN')

    def test_delete_invitation(self):
        self.force_login()
        url = f'/m/invitations/{self.invitation.id}/'
        resp = self.client.delete(url)
        self.assertEqual(resp.status_code, 204)

        self.assertFalse(Invitation.objects.filter(id=self.invitation.id).exists())

    def test_permission_denied(self):
        self.force_login(self.GUEST_USER_ID)
        url = f'/m/invitations/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 403)

        resp = self.client.post(url, {'email': 'test@example.com', 'role': 'MEMBER'})
        self.assertEqual(resp.status_code, 403)



from __future__ import annotations

import uuid
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.utils.functional import cached_property
from saas_base.db import CachedManager
from saas_base.settings import saas_settings
from saas_base.registry import perm_registry
from .group import Group


class Invitation(models.Model):
    class InviteStatus(models.IntegerChoices):
        PENDING = 1, 'pending'
        SENT = 2, 'sent'
        ACCEPTED = 3, 'accepted'
        EXPIRED = 4, 'expired'

    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    tenant = models.ForeignKey(settings.SAAS_TENANT_MODEL, on_delete=models.CASCADE)
    inviter = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, related_name='sent_invitations'
    )
    email = models.EmailField()
    role = models.CharField(max_length=50, blank=True, null=True)
    groups = models.ManyToManyField(Group, blank=True)
    status = models.IntegerField(default=InviteStatus.PENDING, choices=InviteStatus.choices)
    accepted_at = models.DateTimeField(null=True, blank=True)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    class Meta:
        verbose_name = _('invitation')
        verbose_name_plural = _('invitations')
        unique_together = [
            ['tenant', 'email'],
        ]
        ordering = ['-created_at']
        db_table = 'saas_member_invitation'

    def __str__(self):
        return self.email

    def is_expired(self) -> bool:
        duration = timezone.now() - self.created_at
        return duration.days >= saas_settings.INVITATION_EXPIRES_DAYS


class MemberManager(CachedManager['Member']):
    natural_key = ['tenant_id', 'user_id']
    query_select_related = ['user']
    query_prefetch_related = ['groups']

    def get_by_natural_key(self, tenant_id, user_id) -> 'Member':
        return self.get_from_cache_by_natural_key(tenant_id, user_id)


class Member(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    tenant = models.ForeignKey(settings.SAAS_TENANT_MODEL, on_delete=models.CASCADE)
    # TODO: remove null=True
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    role = models.CharField(max_length=50, blank=True, null=True)
    groups = models.ManyToManyField(Group, blank=True)
    permissions = models.JSONField(default=list)
    objects = MemberManager()

    class Meta:
        verbose_name = _('member')
        verbose_name_plural = _('members')
        unique_together = [
            ['tenant', 'user'],
        ]
        ordering = ['-created_at']
        db_table = 'saas_member'

    def __str__(self):
        return str(self.user)

    def natural_key(self):
        return self.tenant_id, self.user_id

    @cached_property
    def group_permissions(self) -> list[str]:
        return list(self.__get_group_permissions())

    @cached_property
    def role_permissions(self) -> list[str]:
        return self.__get_role_permissions()

    @cached_property
    def user_permissions(self) -> list[str]:
        return self.__get_user_permissions()

    def get_all_permissions(self) -> set[str]:
        perms = set()
        perms = perms.union(self.user_permissions)
        perms = perms.union(self.role_permissions)
        perms = perms.union(self.group_permissions)
        return perms

    def __get_group_permissions(self):
        perms = set([])
        queryset = self.groups.values('permissions')
        for data in queryset:
            perms.update(data['permissions'])
        return perms

    def __get_role_permissions(self):
        if self.role:
            return perm_registry.get_role(self.role).permissions
        return []

    def __get_user_permissions(self):
        return self.permissions



# Generated by Django 6.0 on 2025-12-29 06:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('saas_base', '0003_remove_member_is_staff'),
    ]

    operations = [
        migrations.AddField(
            model_name='tenant',
            name='logo',
            field=models.URLField(blank=True, null=True),
        ),
    ]



from django.utils.translation import gettext_lazy as _
from .registry import perm_registry

# Special default role
perm_registry.register_role('OWNER', _('Owner'), _('Full organization access'))
perm_registry.assign_to_role('OWNER', '*')

# User Profile
perm_registry.register_permission(
    key='user.profile.view',
    label=_('View Profile'),
    module='User',
    description=_('Read access to basic profile details (name, avatar, bio)'),
)
perm_registry.register_permission(
    key='user.profile.update',
    label=_('Update Profile'),
    module='User',
    description=_('Write access to update own profile details'),
)

# User Emails
perm_registry.register_permission(
    key='user.email.view',
    label=_('View Emails'),
    module='User',
    description=_('List all registered email addresses'),
)
perm_registry.register_permission(
    key='user.email.manage',
    label=_('Manage Emails'),
    module='User',
    description=_('Add, delete, or verify secondary email addresses'),
)

# User Tenants
perm_registry.register_permission(
    key='user.org.view',
    label=_('View Organizations'),
    module='User',
    description=_('List all organizations you belong to'),
)
perm_registry.register_permission(
    key='user.org.create',
    label=_('Create Organizations'),
    module='User',
    description=_('Create new organizations'),
)
perm_registry.register_permission(
    key='user.or.join',
    label=_('Join Organizations'),
    module='User',
    description=_("Join any organization you're invited to"),
)
perm_registry.register_permission(
    key='user.org.leave',
    label=_('Leave Organizations'),
    module='User',
    description=_('Leave any organization you belong to'),
)

# Identity Management (IAM)
perm_registry.register_permission(
    'iam.member.view',
    _('View Members'),
    module='IAM',
    description=_('List and view member details'),
)
perm_registry.register_permission(
    'iam.member.manage',
    _('Manage Members'),
    module='IAM',
    description=_('Invite, update roles, or remove members'),
)
perm_registry.register_permission(
    'iam.group.view',
    _('View Groups'),
    module='IAM',
    description=_('View tenant groups'),
)
perm_registry.register_permission(
    'iam.group.manage',
    _('Manage Groups'),
    module='IAM',
    description=_('Create or edit tenant-specific groups'),
)
perm_registry.register_permission(
    key='iam.invitation.view',
    label=_('View Invitations'),
    module='IAM',
    description=_('List all pending invitations'),
)
perm_registry.register_permission(
    key='iam.invitation.manage',
    label='Manage Invitations',
    module='IAM',
    description=_('Invite new members to join the organization'),
)

# Organization Management
perm_registry.register_permission(
    'org.info.view',
    _('View Organization'),
    module='Organization',
    description=_('View tenant settings and info'),
)
perm_registry.register_permission(
    'org.info.update',
    _('Update Organization'),
    module='Organization',
    description=_('Update tenant name, logo, etc.'),
)



from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext as _
from rest_framework import serializers
from saas_base.drf.serializers import ModelSerializer, RelatedSerializerField
from saas_base.models import Member, UserEmail, Invitation
from saas_base.registry import perm_registry
from .tenant import TenantSerializer
from .user import SimpleUserSerializer
from .group import GroupSerializer


class InvitationSerializer(ModelSerializer):
    email = serializers.EmailField(required=True)
    role = serializers.CharField(required=False)
    groups = RelatedSerializerField(GroupSerializer, many=True, required=False)

    class Meta:
        model = Invitation
        exclude = ['tenant', 'inviter']
        read_only_fields = ['status', 'accepted_at', 'created_at']
        request_include_fields = ['groups']

    def validate_role(self, role: str):
        if role not in perm_registry.get_role_codes():
            raise serializers.ValidationError(_('Invalid role.'))
        return role

    def validate_email(self, email: str):
        request = self.context['request']
        if Invitation.objects.filter(tenant_id=request.tenant_id, email=email).count():
            raise serializers.ValidationError(_('This email has already been invited.'))

        try:
            user_email = UserEmail.objects.get_by_email(email)
        except ObjectDoesNotExist:
            return email

        if Member.objects.filter(tenant_id=request.tenant_id, user_id=user_email.user_id).count():
            raise serializers.ValidationError(_('This user has already been invited.'))

        return email


class InvitationInfoSerializer(ModelSerializer):
    tenant = TenantSerializer(read_only=True)
    inviter = SimpleUserSerializer(read_only=True)

    class Meta:
        model = Invitation
        fields = '__all__'



from .registry import perm_registry

# 1. Define Core Roles
perm_registry.register_role('ADMIN', 'Administrator', 'Staff management access')
perm_registry.register_role('MEMBER', 'Member', 'Regular access')

# 2. Register Permissions
perm_registry.assign_to_role('ADMIN', 'iam.*')
perm_registry.assign_to_role('ADMIN', 'org.*')
perm_registry.assign_to_role('MEMBER', 'iam.*.view')
perm_registry.assign_to_role('MEMBER', 'org.*.view')



import re
from django.core import mail
from tests.client import FixturesTestCase
from saas_base.models import UserEmail


class TestPasswordAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def setUp(self):
        user = self.get_user()
        UserEmail.objects.create(user=user, email='hi@foo.com', primary=True, verified=True)

    def test_reset_password(self):
        data = {'email': 'hi@foo.com'}
        resp = self.client.post('/m/auth/password/forgot/', data=data)
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(len(mail.outbox), 1)
        msg = mail.outbox[0]
        codes = re.findall(r'Code: (\w{6})', msg.body)

        data = {**data, 'code': codes[0], 'password': 'this is me'}
        resp = self.client.post('/m/auth/password/reset/', data=data)
        self.assertEqual(resp.status_code, 200)
        self.assertIn('next', resp.json())

    def test_wrong_email(self):
        data = {'email': '404@foo.com'}
        resp = self.client.post('/m/auth/password/forgot/', data=data)
        self.assertEqual(resp.status_code, 400)

    def test_wrong_code(self):
        data = {'email': 'hi@foo.com'}
        resp = self.client.post('/m/auth/password/forgot/', data=data)
        self.assertEqual(resp.status_code, 204)
        data = {**data, 'code': 'AAAAAA', 'password': 'this is me'}
        resp = self.client.post('/m/auth/password/reset/', data=data)
        self.assertEqual(resp.status_code, 400)



from saas_base.test import SaasTestCase


class TestCurrentTenantAPI(SaasTestCase):
    def test_fetch_openapi(self):
        resp = self.client.get('/schema/openapi')
        self.assertEqual(resp.status_code, 200)
        self.assertIn('openapi', resp.json())



import typing as t
from .rules import Rule

class TypedSite(t.TypedDict):
    name: str
    url: str
    icon: str
    copyright: str

class TypedProvider(t.TypedDict):
    backend: str
    options: t.Dict[str, t.Any]

class Settings:
    SITE: TypedSite
    TENANT_ID_HEADER: str
    CLIENT_IP_HEADERS: t.Optional[t.List[str]]

    ENABLE_GRAVATAR: bool
    GRAVATAR_PARAMS: t.Dict[str, t.Any]

    DEFAULT_REGION: str
    DEFAULT_FROM_EMAIL: t.Optional[str]

    LOGIN_SECURITY_RULES: t.List[Rule]
    SIGNUP_SECURITY_RULES: t.List[Rule]
    RESET_PASSWORD_SECURITY_RULES: t.List[Rule]
    SIGNUP_REQUEST_CREATE_USER: bool

    INVITATION_LINK: str
    INVITATION_EXPIRES_DAYS: int

    # internal properties
    settings_key: str
    user_settings: t.Dict[str, t.Any]
    _cached_attrs: t.Set[str]

    def __init__(
        self,
        settings_key: t.Optional[str] = 'SAAS',
        user_settings: t.Optional[t.Dict[str, t.Any]] = None,
        defaults: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> 'Settings': ...
    def reload(self, value: t.Optional[t.Dict[str, t.Any]] = None): ...
    def listen_setting_changed(self, setting: str, **kwargs): ...

def perform_import_provider(data: TypedProvider): ...
def perform_import(val: t.Union[t.List[TypedProvider], t.Dict[str, TypedProvider], TypedProvider]): ...

saas_settings: Settings



from unittest.mock import Mock
from saas_base.ipware import get_client_ip
from saas_base.settings import saas_settings
from rest_framework.settings import api_settings


def test_get_client_ip_remote_addr():
    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {}

    assert get_client_ip(request) == '127.0.0.1'


def test_get_client_ip_x_forwarded_for():
    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'X-Forwarded-For': '10.0.0.1'}

    assert get_client_ip(request) == '10.0.0.1'


def test_get_client_ip_custom_headers():
    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'Custom-Header': '10.0.0.2'}

    assert get_client_ip(request, ip_headers=['Custom-Header']) == '10.0.0.2'


def test_get_client_ip_settings_headers(monkeypatch):
    monkeypatch.setattr(saas_settings, 'CLIENT_IP_HEADERS', ['My-IP'])

    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'My-IP': '10.0.0.3'}

    assert get_client_ip(request) == '10.0.0.3'


def test_get_client_ip_proxied_no_num_proxies(monkeypatch):
    monkeypatch.setattr(api_settings, 'NUM_PROXIES', None)

    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'X-Forwarded-For': '10.0.0.1, 10.0.0.2'}

    # When NUM_PROXIES is None, it returns the joined string without spaces
    assert get_client_ip(request) == '10.0.0.1,10.0.0.2'


def test_get_client_ip_proxied_with_num_proxies(monkeypatch):
    monkeypatch.setattr(api_settings, 'NUM_PROXIES', 1)

    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'X-Forwarded-For': '10.0.0.1, 10.0.0.2'}

    # With NUM_PROXIES=1, it should take the last one
    assert get_client_ip(request) == '10.0.0.2'


def test_get_client_ip_proxied_with_num_proxies_2(monkeypatch):
    monkeypatch.setattr(api_settings, 'NUM_PROXIES', 2)

    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'X-Forwarded-For': '10.0.0.1, 10.0.0.2, 10.0.0.3'}

    # With NUM_PROXIES=2, it should take the second to last
    assert get_client_ip(request) == '10.0.0.2'


def test_get_client_ip_proxied_zero_proxies(monkeypatch):
    monkeypatch.setattr(api_settings, 'NUM_PROXIES', 0)

    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    request.headers = {'X-Forwarded-For': '10.0.0.1, 10.0.0.2'}

    # With NUM_PROXIES=0, it should return REMOTE_ADDR
    assert get_client_ip(request) == '127.0.0.1'


def test_get_client_ip_priority():
    request = Mock()
    request.META = {'REMOTE_ADDR': '127.0.0.1'}
    # X-Forwarded-For is before X-Real-IP in default list
    request.headers = {
        'X-Real-IP': '10.0.0.2',
        'X-Forwarded-For': '10.0.0.1',
    }

    assert get_client_ip(request) == '10.0.0.1'



from .registry import perm_registry

perm_registry.register_scope(
    key='openid',
    label='OpenID',
    permissions=[],  # Mandatory for OIDC, used as an auth flag
    is_oidc=True,
    description='Standard OIDC identifier',
)
perm_registry.register_scope(
    key='profile',
    label='Profile Information',
    permissions=['user.profile.view'],
    is_oidc=True,
    description='Access to your name, nickname, and picture',
)
perm_registry.register_scope(
    key='email',
    label='Email Address',
    permissions=['user.email.view'],
    is_oidc=True,
    description='Access to your primary and secondary emails',
)

perm_registry.register_scope(
    key='user:read',
    label='Read User Account',
    permissions=['user.*.view'],
    description='Read-only access to all personal account data',
)

perm_registry.register_scope(
    key='user:write',
    label='Manage User Account',
    permissions=[
        'user.profile.update',
        'user.*.manage',
    ],
    description='Full write access to profile and secondary security settings',
)

perm_registry.register_scope(
    key='user',
    label='Full Personal Access',
    permissions=['user.*'],
    description='Complete control over personal global resources',
)

perm_registry.register_scope(
    key='org:read',
    label='Read Organization',
    permissions=['org.*.view'],
    description='Read-only access to organization settings',
)

perm_registry.register_scope(
    key='org:write',
    label='Manage Organization',
    permissions=['org.*.manage'],
    description='Full write access to organization settings',
)

perm_registry.register_scope(
    key='org',
    label='Full Organization Access',
    permissions=['org.*'],
    description='Complete control over organization resources',
)



import uuid
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from ..db import CachedManager


class GroupManager(CachedManager['Group']):
    natural_key = ['tenant_id', 'name']
    query_select_related = ['tenant']

    def get_by_natural_key(self, tenant_id, name):
        return self.get_from_cache_by_natural_key(tenant_id, name)


class Group(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    tenant = models.ForeignKey(settings.SAAS_TENANT_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=100, db_index=True)
    permissions = models.JSONField(default=list)
    managed = models.BooleanField(default=False)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    objects = GroupManager()

    class Meta:
        verbose_name = _('group')
        verbose_name_plural = _('groups')
        unique_together = [['tenant', 'name']]
        db_table = 'saas_group'
        ordering = ['created_at']

    def __str__(self):
        return self.name

    def natural_key(self):
        return self.tenant_id, self.name



from django.apps import AppConfig
from django.conf import settings

if not hasattr(settings, 'SAAS_TENANT_MODEL'):
    setattr(settings, 'SAAS_TENANT_MODEL', 'saas_base.Tenant')


class CoreConfig(AppConfig):
    name = 'saas_base'
    verbose_name = 'SaaS'

    def ready(self):
        from ._ready import setup

        setup()



from django.contrib import admin
from .models import (
    Group,
    Member,
    UserEmail,
    Tenant,
    UserProfile,
    Invitation,
)


@admin.register(Tenant)
class TenantAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'slug', 'owner', 'expires_at', 'created_at']


@admin.register(Group)
class GroupAdmin(admin.ModelAdmin):
    list_display = ['id', 'tenant', 'name', 'managed', 'created_at']


@admin.register(Member)
class MemberAdmin(admin.ModelAdmin):
    list_display = ['id', 'tenant', 'user', 'created_at']


@admin.register(UserEmail)
class UserEmailAdmin(admin.ModelAdmin):
    list_display = ['id', 'email', 'primary', 'verified', 'created_at']


@admin.register(UserProfile)
class UserProfileAdmin(admin.ModelAdmin):
    list_display = ['user', 'region', 'locale', 'timezone']


@admin.register(Invitation)
class InvitationAdmin(admin.ModelAdmin):
    list_display = ['tenant', 'email', 'role', 'created_at']



from django.utils.translation import gettext as _
from django.db.models import OuterRef, Subquery, Count
from rest_framework.request import Request
from rest_framework.exceptions import PermissionDenied
from rest_framework.mixins import (
    ListModelMixin,
    DestroyModelMixin,
)
from saas_base.drf.filters import CurrentUserFilter, IncludeFilter
from saas_base.drf.views import AuthenticatedEndpoint
from saas_base.drf.decorators import resource_permission
from saas_base.models import Member
from saas_base.serializers.user_tenant import UserTenantSerializer

__all__ = [
    'UserTenantListEndpoint',
    'UserTenantItemEndpoint',
]


class UserTenantListEndpoint(ListModelMixin, AuthenticatedEndpoint):
    serializer_class = UserTenantSerializer
    queryset = Member.objects.select_related('tenant').all()
    pagination_class = None

    filter_backends = [CurrentUserFilter, IncludeFilter]
    include_prefetch_related_fields = ['groups']
    include_annotate_fields = ['member_count']

    def annotate_queryset(self, queryset, terms):
        if 'member_count' in terms:
            member_count_subquery = (
                Member.objects.filter(tenant=OuterRef('tenant'))
                .values('tenant')
                .annotate(count=Count('id'))
                .values('count')
            )
            queryset = queryset.annotate(member_count=Subquery(member_count_subquery))
            print(queryset.query)
        return queryset

    @resource_permission('user.org.view')
    def get(self, request: Request, *args, **kwargs):
        """List all the current user's tenants."""
        return self.list(request, *args, **kwargs)


class UserTenantItemEndpoint(DestroyModelMixin, AuthenticatedEndpoint):
    queryset = Member.objects.all()
    filter_backends = [CurrentUserFilter]
    lookup_field = 'tenant_id'

    @resource_permission('user.org.leave')
    def delete(self, request: Request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)

    def perform_destroy(self, instance: Member):
        if instance.tenant.owner_id == self.request.user.pk:
            raise PermissionDenied(_('Cannot leave your own tenant.'))
        instance.delete()



from django.core import mail
from saas_base.models import UserEmail
from tests.client import FixturesTestCase


class TestEmailAPI(FixturesTestCase):
    user_id = FixturesTestCase.GUEST_USER_ID

    def test_list_emails(self):
        self.force_login()

        url = '/m/user/emails/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), [])

        for i in range(5):
            UserEmail.objects.create(user=self.user, email=f'demo-{self.user_id}-{i}@example.com')

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 5)

    def test_add_email(self):
        self.force_login()
        data1 = {'email': 'demo-1-1@example.com'}
        resp = self.client.post('/m/user/emails/add/request/', data1)
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(len(mail.outbox), 1)

        data2 = {**data1, 'code': self.get_mail_auth_code()}
        resp = self.client.post('/m/user/emails/add/confirm/', data=data2)
        self.assertEqual(resp.status_code, 200)
        item = resp.json()

        url = '/m/user/emails/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), [item])

    def test_set_primary_email(self):
        self.force_login()

        obj = UserEmail.objects.create(
            user=self.user,
            email='demo-1-2@example.com',
            verified=True,
        )
        url = f'/m/user/emails/{obj.pk}/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertFalse(resp.json()['primary'])

        resp = self.client.patch(url, data={'primary': True})
        self.assertEqual(resp.status_code, 200)
        self.assertTrue(resp.json()['primary'])

    def test_delete_email(self):
        self.force_login()
        obj = UserEmail.objects.create(
            user=self.user,
            email='demo-1-2@example.com',
            verified=True,
        )
        url = f'/m/user/emails/{obj.pk}/'
        resp = self.client.delete(url)
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(UserEmail.objects.filter(user=self.user).count(), 0)



from django.urls import path
from ..endpoints.user_tenants import (
    UserTenantListEndpoint,
    UserTenantItemEndpoint,
)

urlpatterns = [
    path('', UserTenantListEndpoint.as_view()),
    path('<tenant_id>/', UserTenantItemEndpoint.as_view()),
]



from django.db import models
from django.conf import settings


class UserProfile(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name='profile',
    )
    bio = models.TextField(max_length=500, blank=True)
    birth_date = models.DateField(null=True, blank=True)
    website = models.URLField(blank=True, null=True)
    avatar_url = models.URLField(blank=True, null=True)
    region = models.CharField(blank=True, null=True, max_length=4)
    locale = models.CharField(blank=True, null=True, max_length=10)
    timezone = models.CharField(max_length=50, default='UTC', help_text="e.g., 'America/New_York'")

    class Meta:
        db_table = 'saas_user_profile'



from django.urls import path, include
from ..endpoints.permissions import PermissionListEndpoint
from ..endpoints.roles import RoleListEndpoint
from ..endpoints.tenant import SelectedTenantEndpoint

urlpatterns = [
    path('permissions/', PermissionListEndpoint.as_view()),
    path('roles/', RoleListEndpoint.as_view()),
    path('tenant/', SelectedTenantEndpoint.as_view()),
    path('user/', include('saas_base.api_urls.user')),
    path('user/emails/', include('saas_base.api_urls.user_emails')),
    path('user/tenants/', include('saas_base.api_urls.user_tenants')),
    path('user/invitations/', include('saas_base.api_urls.user_invitations')),
    path('tenants/', include('saas_base.api_urls.tenants')),
    path('groups/', include('saas_base.api_urls.groups')),
    path('members/', include('saas_base.api_urls.members')),
    path('invitations/', include('saas_base.api_urls.invitations')),
    path('auth/', include('saas_base.api_urls.auth')),
]



import functools


def resource_permission(name):
    """A decorator that adding required resource permissions to the view function.

    .. code-block:: python

        class TestView(TenantEndpoint):
            @resource_permissions('org.settings.update')
            def post(self, request, *args, **kwargs):
                ...
    """

    def decorated(func):
        func._required_permission = name

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapped

    return decorated



import typing as t

from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.utils.translation import gettext_lazy as _
from django.contrib.auth import password_validation, get_user_model
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.contrib.auth.models import User
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from .email_code import (
    EmailCode,
    EmailCodeRequestSerializer,
    EmailCodeConfirmSerializer,
    NewUserEmailMixin,
    RetrieveUserEmailMixin,
)
from saas_base.models import UserEmail, Invitation, Member

SIGNUP_CODE = 'saas:signup_code'

AUTH_ERRORS = {
    'exist_username': _('This username is already associated with an existing account.'),
}


class SignupRequestCodeSerializer(NewUserEmailMixin, EmailCodeRequestSerializer):
    CACHE_PREFIX = SIGNUP_CODE

    def create(self, validated_data) -> EmailCode:
        code = self.save_auth_code(1)
        email = validated_data['email']
        return EmailCode(email, code)


class SignupCreateUserSerializer(NewUserEmailMixin, EmailCodeRequestSerializer):
    CACHE_PREFIX = SIGNUP_CODE
    username = serializers.CharField(required=True, validators=[UnicodeUsernameValidator()])
    password = serializers.CharField(required=True)

    def validate_username(self, username: str):
        cls: t.Type[User] = get_user_model()
        try:
            cls.objects.get(username=username)
            raise ValidationError(AUTH_ERRORS['exist_username'])
        except cls.DoesNotExist:
            return username

    def validate_password(self, raw_password: str):
        user = User(
            username=self.initial_data['username'],
            email=self.initial_data['email'],
        )
        password_validation.validate_password(raw_password, user)
        return raw_password

    def create(self, validated_data) -> EmailCode:
        username = validated_data['username']
        email = validated_data['email']
        password = validated_data['password']
        cls: t.Type[User] = get_user_model()
        with transaction.atomic():
            user = cls.objects.create_user(
                username=username,
                email=email,
                password=password,
                is_active=False,
            )
            UserEmail.objects.create(user=user, email=email, primary=True, verified=False)

        code = self.save_auth_code(1)
        return EmailCode(email, code, user)


class SignupConfirmCodeSerializer(RetrieveUserEmailMixin, EmailCodeConfirmSerializer):
    CACHE_PREFIX = SIGNUP_CODE

    def create(self, validated_data) -> User:
        user_email = validated_data['email']
        with transaction.atomic():
            user_email.verified = True
            user = user_email.user
            user.is_active = True
            user_email.save()
            user.save()
        return user


class SignupConfirmPasswordSerializer(EmailCodeConfirmSerializer):
    CACHE_PREFIX = SIGNUP_CODE
    username = serializers.CharField(required=True, validators=[UnicodeUsernameValidator()])
    password = serializers.CharField(required=True)
    email = serializers.EmailField(required=True)
    code = serializers.CharField(required=True, max_length=6)

    def validate_username(self, username: str):
        cls: t.Type[User] = get_user_model()
        try:
            cls.objects.get(username=username)
            raise ValidationError(AUTH_ERRORS['exist_username'])
        except ObjectDoesNotExist:
            return username

    def validate_password(self, raw_password: str):
        user = User(
            username=self.initial_data['username'],
            email=self.initial_data['email'],
        )
        password_validation.validate_password(raw_password, user)
        return raw_password

    def create(self, validated_data) -> User:
        username = validated_data['username']
        email = validated_data['email']
        password = validated_data['password']
        cls: t.Type[User] = get_user_model()
        with transaction.atomic():
            user = cls.objects.create_user(
                username=username,
                email=email,
                password=password,
                is_active=True,
            )
            UserEmail.objects.create(
                user=user,
                email=email,
                primary=True,
                verified=True,
            )
        return user


class SignupWithInvitationSerializer(serializers.Serializer):
    username = serializers.CharField(required=True, validators=[UnicodeUsernameValidator()])
    password = serializers.CharField(required=True)

    def validate_username(self, username: str):
        cls: t.Type[User] = get_user_model()
        try:
            cls.objects.get(username=username)
            raise ValidationError(AUTH_ERRORS['exist_username'])
        except ObjectDoesNotExist:
            return username

    def validate_password(self, raw_password: str):
        invitation = self.context['invitation']
        user = User(
            username=self.initial_data['username'],
            email=invitation.email,
        )
        password_validation.validate_password(raw_password, user)
        return raw_password

    def create(self, validated_data) -> User:
        username = validated_data['username']
        password = validated_data['password']
        invitation = self.context['invitation']

        cls: t.Type[User] = get_user_model()
        with transaction.atomic():
            user = cls.objects.create_user(
                username=username,
                email=invitation.email,
                password=password,
                is_active=True,
            )
            UserEmail.objects.create(
                user=user,
                email=invitation.email,
                primary=True,
                verified=True,
            )
            member = Member.objects.create(
                tenant_id=invitation.tenant_id,
                user=user,
                role=invitation.role,
            )
            member.groups.set(invitation.groups.all())
            invitation.status = Invitation.InviteStatus.ACCEPTED
            invitation.save()
        return user



from saas_base.test import SaasTestCase
from saas_base.drf.serializers import FlattenModelSerializer
from saas_base.models import Member
from saas_base.serializers.user import SimpleUserSerializer


class MemberSerializer(FlattenModelSerializer):
    user = SimpleUserSerializer(read_only=True)

    class Meta:
        model = Member
        fields = '__all__'
        flatten_fields = ['user']


class TestModelSerializer(SaasTestCase):
    user_id = SaasTestCase.OWNER_USER_ID

    def test_flatten_fields(self):
        user = self.get_user()
        user.first_name = 'Django'
        user.save()
        member = Member.objects.get(user=user, tenant_id=self.tenant_id)
        serializer = MemberSerializer(member)
        self.assertEqual(serializer.data['name'], 'Django')



from django.utils.translation import gettext as _
from django.contrib.auth import get_user_model
from django.contrib.auth import password_validation
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from saas_base.drf.serializers import FlattenModelSerializer
from saas_base.models import UserProfile
from saas_base.settings import saas_settings
from saas_base.utils.gravatar import gen_gravatar_url

UserModel = get_user_model()


class UserProfileSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserProfile
        exclude = ('user',)


class UserSerializer(FlattenModelSerializer):
    name = serializers.CharField(source='get_full_name', read_only=True)
    profile = UserProfileSerializer()

    class Meta:
        model = UserModel
        exclude = ['password', 'groups', 'user_permissions']
        flatten_fields = ['profile']

    def to_representation(self, instance):
        data = super().to_representation(instance)

        # use gravatar if enabled
        if not data.get('avatar_url') and saas_settings.ENABLE_GRAVATAR:
            name = instance.get_full_name() or instance.username
            avatar_url = gen_gravatar_url(instance.email, name, **saas_settings.GRAVATAR_PARAMS)
            data['avatar_url'] = avatar_url
        return data


class SimpleUserSerializer(serializers.ModelSerializer):
    name = serializers.CharField(source='get_full_name', read_only=True)
    username = serializers.CharField(source='get_username', read_only=True)

    class Meta:
        model = UserModel
        fields = ['id', 'username', 'name', 'is_active', 'is_staff']


class UserPasswordSerializer(serializers.Serializer):
    old_password = serializers.CharField(required=True)
    password = serializers.CharField(required=True)
    confirm_password = serializers.CharField(required=True)

    def validate_old_password(self, value):
        user = self.instance
        if not user.check_password(value):
            raise ValidationError(_('Password incorrect.'))
        return value

    def validate_password(self, raw_password: str):
        if self.initial_data['confirm_password'] != raw_password:
            raise ValidationError(_('Password does not match.'))
        password_validation.validate_password(raw_password)
        return raw_password

    def update(self, user, validated_data):
        raw_password = validated_data['password']
        user.set_password(raw_password)
        user.save()
        return user



from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
from django.utils.translation import gettext as _
from rest_framework.permissions import BasePermission, SAFE_METHODS
from rest_framework.request import Request
from rest_framework.exceptions import PermissionDenied
from saas_base.models import get_tenant_model, get_cached_tenant, Member
from saas_base.registry import perm_registry
from saas_base.registry.checker import has_permission, get_view_permission

__all__ = [
    'IsTenantOwner',
    'IsTenantOwnerOrReadOnly',
    'IsTenantActive',
    'IsTenantActiveOrReadOnly',
    'HasResourcePermission',
    'HasResourcePermissionOrReadOnly',
]

TenantModel = get_tenant_model()


class BaseTenantPermission(BasePermission):
    tenant_id_field = 'tenant_id'

    @staticmethod
    def check_tenant_permission(request, view, tenant_id):
        return True

    def has_permission(self, request: Request, view):
        if not request.user.is_authenticated:
            return False
        tenant_id = getattr(request, 'tenant_id', None)
        return self.check_tenant_permission(request, view, tenant_id)

    def has_object_permission(self, request, view, obj):
        # permission checked above
        tenant_id = getattr(request, 'tenant_id', None)
        if tenant_id:
            return True

        if hasattr(view, 'object_tenant_id_field'):
            tenant_id_field = view.object_tenant_id_field
        else:
            tenant_id_field = getattr(view, 'tenant_id_field', self.tenant_id_field)

        if tenant_id_field is None:
            return True

        tenant_id = getattr(obj, tenant_id_field, None)
        return self.check_tenant_permission(request, view, tenant_id)


class IsTenantOwner(BaseTenantPermission):
    """The authenticated user is the tenant owner."""

    @staticmethod
    def check_tenant_permission(request, view, tenant_id):
        if not tenant_id:
            return False

        member = get_requested_member(request, tenant_id)
        if member:
            # OWNER is a system role code
            return member.role == 'OWNER'

        tenant = get_cached_tenant(tenant_id, request._request)
        if not tenant:
            return False

        # this user is primary owner
        return request.user.pk == tenant.owner_id


class IsTenantOwnerOrReadOnly(IsTenantOwner):
    """The authenticated user is the tenant owner, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS:
            return True
        return super().has_object_permission(request, view, obj)


class IsTenantActive(BaseTenantPermission):
    """The requested tenant is not expired."""

    @staticmethod
    def check_tenant_permission(request, view, tenant_id):
        tenant = get_cached_tenant(tenant_id, request._request)
        if not tenant:
            return False

        if not tenant.expires_at:
            return True

        if tenant.expires_at < timezone.now():
            raise PermissionDenied(_('This tenant is expired.'))
        return True


class IsTenantActiveOrReadOnly(IsTenantActive):
    """The requested tenant is not expired, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS:
            return True
        return super().has_object_permission(request, view, obj)


class HasResourcePermission(BasePermission):
    """The authenticated user is a member of the tenant, and the user
    has the given resource permission.
    """

    tenant_id_field = 'tenant_id'

    def has_permission(self, request: Request, view):
        required_perm = get_view_permission(view, request.method)
        if not required_perm:
            return True

        if not request.user.is_authenticated:
            return False

        # filter scopes permissions
        if request.auth:
            scopes_perms = self.get_scopes_permissions(request.auth)
            if not has_permission(required_perm, scopes_perms):
                return False

        # this is a tenant related permission
        if not required_perm.startswith('user.'):
            tenant_id = getattr(request, 'tenant_id', None)
            return self.check_tenant_permission(request, required_perm, tenant_id)

        # user related permission has been checked above by token scopes permissions
        # if there is no `request.auth`, it is a session based authentication
        return True

    def has_object_permission(self, request, view, obj):
        # permission checked above
        tenant_id = getattr(request, 'tenant_id', None)
        if tenant_id:
            return True

        if hasattr(view, 'object_tenant_id_field'):
            tenant_id_field = view.object_tenant_id_field
        else:
            tenant_id_field = getattr(view, 'tenant_id_field', self.tenant_id_field)

        if tenant_id_field is None:
            return True

        # it is not related with tenant
        if not hasattr(obj, tenant_id_field):
            return True

        tenant_id = getattr(obj, tenant_id_field, None)
        return self.check_tenant_permission(request, view, tenant_id)

    @staticmethod
    def get_scopes_permissions(token):
        if hasattr(token, 'scope') and token.scope:
            scopes = token.scope.split(' ')
        elif hasattr(token, 'scopes'):
            scopes = token.scopes
        elif hasattr(token, 'get_scopes'):
            scopes = token.get_scopes()
        else:
            scopes = []

        if not scopes:
            return

        return perm_registry.get_permissions_for_scopes(scopes)

    @classmethod
    def check_tenant_permission(cls, request: Request, required_perm: str, tenant_id):
        if not tenant_id:
            return False

        tenant = get_cached_tenant(tenant_id, request._request)
        if not tenant:
            return False

        # Tenant owner has all permissions
        if tenant.owner_id == request.user.pk:
            return True

        member = get_requested_member(request, tenant_id)
        if not member:
            return False

        assigned_perms = member.get_all_permissions()
        return has_permission(required_perm, assigned_perms)


class HasResourcePermissionOrReadOnly(HasResourcePermission):
    """The authenticated user has the tenant permission, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS:
            return True
        return super().has_object_permission(request, view, obj)


def get_requested_member(request: Request, tenant_id):
    member = getattr(request, 'member', None)
    if member:
        return member

    try:
        return Member.objects.get_from_cache_by_natural_key(tenant_id, request.user.pk)
    except ObjectDoesNotExist:
        return None



from .tenant import AbstractTenant, Tenant, get_tenant_model, get_cached_tenant
from .group import Group
from .member import Member, Invitation
from .user_profile import UserProfile
from .user_email import UserEmail

__all__ = [
    'AbstractTenant',
    'Tenant',
    'get_tenant_model',
    'get_cached_tenant',
    'Group',
    'Member',
    'Invitation',
    'UserEmail',
    'UserProfile',
]



import typing as t
import hashlib
import logging
import uuid

from django.db.models import Model, Manager, UUIDField, QuerySet
from django.db.models.signals import (
    class_prepared,
    post_save,
    post_delete,
    m2m_changed,
)
from .cache import cache

__all__ = ['CachedManager']

logger = logging.getLogger(__name__)

M = t.TypeVar('M', bound=Model)


class CachedManager(Manager, t.Generic[M]):
    db_cache = cache
    natural_key: list[str] = []
    cache_ttl: int = 300
    cache_version = 2
    query_select_related: list[str] = []
    query_prefetch_related: list[str] = []

    def contribute_to_class(self, model: t.Type[M], name: str):
        super().contribute_to_class(model, name)
        class_prepared.connect(self.__manage_cache, sender=model)

    def _filter_related_queryset(self, queryset: QuerySet[M]) -> QuerySet[M]:
        if self.query_select_related:
            queryset = queryset.select_related(*self.query_select_related)
        if self.query_prefetch_related:
            queryset = queryset.prefetch_related(*self.query_prefetch_related)
        return queryset

    def get_many_from_cache(self, pk_set: list[t.Any]) -> dict[t.Any, M]:
        key_map = {self.__get_lookup_cache_key(pk=pk): pk for pk in pk_set}
        results = self.db_cache.get_many(key_map.keys(), version=self.cache_version)

        found = {}
        missed = []
        for key in key_map:
            if key not in results:
                missed.append(key_map[key])
            else:
                found[key_map[key]] = results[key]

        if missed:
            queryset = self._filter_related_queryset(self.filter(pk__in=missed))
            to_delete = []
            to_save = {}
            to_recover = []
            for instance in queryset:
                resolved = self.__resolve_instance_caches(instance, False)
                to_delete.extend(resolved['delete'])
                to_save.update(resolved['save'])
                to_recover.append(resolved['recover_state'])

            if to_delete:
                self.db_cache.delete_many(to_delete, version=self.cache_version)
            if to_save:
                self.db_cache.set_many(to_save, timeout=self.cache_ttl, version=self.cache_version)

            for instance, db in to_recover:
                instance._state.db = db
                found[instance.pk] = instance
        return found

    def get_from_cache_by_pk(self, pk: t.Any) -> M:
        key = self.__get_lookup_cache_key(pk=pk)
        instance = self.__get_from_cache_or_raise(key)
        if instance:
            instance._state.db = self.db
            return instance

        try:
            instance: M = self._filter_related_queryset(self).get(pk=pk)
        except self.model.DoesNotExist:
            self.__set_not_exist_cache(key)
            raise self.model.DoesNotExist

        self.__save_db_cache(instance)
        return instance

    def get_from_cache_by_natural_key(self, *args) -> M:
        kwargs = dict(zip(self.natural_key, args))
        key = self.__get_lookup_cache_key(**kwargs)
        pk_val = self.__get_from_cache_or_raise(key)
        if pk_val:
            return self.get_from_cache_by_pk(pk_val)

        try:
            instance = self._filter_related_queryset(self).get(**kwargs)
        except self.model.DoesNotExist:
            self.__set_not_exist_cache(key)
            raise self.model.DoesNotExist

        self.__save_db_cache(instance)
        return instance

    def purge(self, pk_value):
        key = self.__get_lookup_cache_key(pk=pk_value)
        self.db_cache.delete(key, version=self.cache_version)

    def purge_many(self, pk_values):
        keys = [self.__get_lookup_cache_key(pk=pk) for pk in pk_values]
        self.db_cache.delete_many(keys, version=self.cache_version)

    def __manage_cache(self, sender, **kwargs):
        uid = f'{sender._meta.label}.cached_manager'
        post_save.connect(self.__post_save, sender=sender, weak=False, dispatch_uid=f'{uid}_save')
        post_delete.connect(self.__post_delete, sender=sender, weak=False, dispatch_uid=f'{uid}_delete')

    def setup_related_cache_invalidation(self):
        main_model = self.model
        for field_name in self.query_select_related:
            field = main_model._meta.get_field(field_name)
            self.__invalidate_related_cache(main_model, field.related_model, field_name)

        for field_name in self.query_prefetch_related:
            field = main_model._meta.get_field(field_name)
            self.__invalidate_related_cache(main_model, field.related_model, field_name)
            if field.many_to_many:
                self.__invalidate_m2m_cache(main_model, field)

    def __invalidate_related_cache(self, main_model, related_model, filter_key):
        def _handler(sender, instance, **kwargs):
            pk_values = main_model.objects.filter(**{filter_key: instance}).values_list('pk', flat=True)
            self.purge_many(pk_values)

        dispatch_uid = f'{main_model._meta.label}__{related_model._meta.label}.cached_manager'
        post_save.connect(_handler, sender=related_model, weak=False, dispatch_uid=f'{dispatch_uid}_save')
        post_delete.connect(_handler, sender=related_model, weak=False, dispatch_uid=f'{dispatch_uid}_delete')

    def __invalidate_m2m_cache(self, main_model, field):
        filter_key = field.remote_field.name

        def m2m_handler(sender, instance, action, **kwargs):
            if action in ['post_add', 'post_remove', 'post_clear']:
                if isinstance(instance, main_model):
                    self.purge(instance.pk)
                else:
                    pk_values = main_model.objects.filter(**{filter_key: instance}).values_list('pk', flat=True)
                    self.purge_many(pk_values)

        m2m_changed.connect(m2m_handler, sender=field.remote_field.through, weak=False)

    def __get_from_cache_or_raise(self, key: str):
        value = self.db_cache.get(key, version=self.cache_version)
        if value == '__none__':
            raise self.model.DoesNotExist
        return value

    def __set_not_exist_cache(self, key: str):
        self.db_cache.set(
            key=key,
            value='__none__',
            timeout=self.cache_ttl,
            version=self.cache_version,
        )

    def __get_lookup_cache_key(self, **kwargs) -> str:
        key = make_key(self.model, kwargs)
        return f'db:{self.model._meta.db_table}:{key}'

    def __get_natural_cache_key(self, instance: M) -> str:
        natural_fields = {key: value_for_field(instance, key) for key in self.natural_key}
        return self.__get_lookup_cache_key(**natural_fields)

    def __post_save(self, instance, **kwargs):
        self.__save_db_cache(instance, **kwargs)
        # transaction.on_commit(lambda: self.__save_db_cache(instance, **kwargs))

    def __resolve_instance_caches(self, instance, created: bool):
        if not created:
            _natural_cache_key = self.db_cache.get(
                self.__get_lookup_cache_key(__track=instance.pk),
                version=self.cache_version,
            )
        else:
            _natural_cache_key = None

        to_delete = []
        natural_cache_key = self.__get_natural_cache_key(instance)
        if _natural_cache_key and _natural_cache_key != natural_cache_key:
            to_delete.append(_natural_cache_key)

        # natural key and pk mapping
        to_save = {
            natural_cache_key: instance.pk,
            self.__get_lookup_cache_key(__track=instance.pk): natural_cache_key,
        }

        # Ensure we don't serialize the database into the cache
        db = instance._state.db
        instance._state.db = None

        instance_key = self.__get_lookup_cache_key(pk=instance.pk)
        to_save[instance_key] = instance
        return {'save': to_save, 'delete': to_delete, 'recover_state': (instance, db)}

    def __save_db_cache(self, instance, **kwargs):
        resolved = self.__resolve_instance_caches(instance, kwargs.get('created', False))

        to_delete = resolved['delete']
        if to_delete:
            self.db_cache.delete_many(to_delete, version=self.cache_version)

        to_save = resolved['save']
        if to_save:
            self.db_cache.set_many(to_save, timeout=self.cache_ttl, version=self.cache_version)

        # recover database on instance
        resolved['recover_state'][0]._state.db = resolved['recover_state'][1]

    def __post_delete(self, instance, **kwargs):
        to_delete = [
            self.__get_lookup_cache_key(pk=instance.pk),
            self.__get_natural_cache_key(instance),
            self.__get_lookup_cache_key(__track=instance.pk),
        ]
        self.db_cache.delete_many(to_delete, version=self.cache_version)


def make_key(cls: Model, kwargs: t.Mapping[str, t.Union[Model, int, str]]) -> str:
    fields = []
    for k, v in sorted(kwargs.items()):
        if k == 'pk':
            # convert pk to its real name
            k = str(cls._meta.pk.name)
            if isinstance(v, str) and isinstance(cls._meta.pk, UUIDField):
                v = v.replace('-', '')
        if isinstance(v, Model):
            v = v.pk
        if isinstance(v, uuid.UUID):
            v = v.hex
        fields.append(f'{k}={v}')
    return hashlib.md5('&'.join(fields).encode('utf-8')).hexdigest()


def value_for_field(instance: M, key: str) -> t.Any:
    field = instance._meta.get_field(key)
    return getattr(instance, field.attname)



import typing as t
from django.db.models import QuerySet
from django.core.cache import cache
from django.conf import settings
from rest_framework.generics import GenericAPIView, get_object_or_404
from rest_framework.settings import api_settings
from rest_framework.permissions import IsAuthenticated
from .errors import BadRequest
from .permissions import IsTenantActiveOrReadOnly, HasResourcePermission
from .filters import TenantIdFilter


__all__ = [
    'Endpoint',
    'AuthenticatedEndpoint',
    'TenantEndpoint',
]


class Endpoint(GenericAPIView):
    permission_classes = [HasResourcePermission] + api_settings.DEFAULT_PERMISSION_CLASSES
    resource_name: t.Optional[str] = None
    resource_action: t.Optional[str] = None

    @staticmethod
    def get_object_or_404(queryset: QuerySet, **kwargs):
        return get_object_or_404(queryset, **kwargs)

    def prevent_duplicate_request(self, suffix: t.Union[str, int], timeout: int = 120, force: bool = False):
        if getattr(settings, 'TESTING', False) and not force:
            return

        path = self.request.path
        key = f'request:{path}:{suffix}'
        if cache.get(key):
            raise BadRequest('Duplicate request')
        cache.set(key, 1, timeout=timeout)


class AuthenticatedEndpoint(Endpoint):
    permission_classes = [IsAuthenticated, HasResourcePermission] + api_settings.DEFAULT_PERMISSION_CLASSES


class TenantEndpoint(Endpoint):
    permission_classes = [
        IsAuthenticated,
        IsTenantActiveOrReadOnly,
        HasResourcePermission,
    ] + api_settings.DEFAULT_PERMISSION_CLASSES
    filter_backends = [TenantIdFilter]
    resource_name = 'tenant'

    def get_tenant_id(self):
        tenant_id = getattr(self.request, 'tenant_id', None)
        if not tenant_id:
            raise BadRequest('Missing Tenant ID')
        return tenant_id



from rest_framework import serializers
from saas_base.drf.serializers import FlattenModelSerializer
from saas_base.models import Member
from saas_base.registry import perm_registry
from saas_base.registry.checker import has_permission
from .tenant import TenantSerializer


class UserTenantSerializer(FlattenModelSerializer):
    tenant = TenantSerializer(required=True)
    groups = serializers.SlugRelatedField(many=True, read_only=True, slug_field='name')
    permissions = serializers.SerializerMethodField()
    member_count = serializers.IntegerField(read_only=True)

    class Meta:
        model = Member
        exclude = ['user', 'id', 'created_at']
        flatten_fields = ['tenant']
        request_include_fields = ['groups', 'member_count']

    def get_permissions(self, obj: Member):
        assigned_patterns = obj.get_all_permissions()

        all_registered_perms = perm_registry.get_permission_list()
        if obj.tenant.owner_id == obj.user_id:
            return [p.key for p in all_registered_perms]

        resolved_keys = [p.key for p in all_registered_perms if has_permission(p.key, assigned_patterns)]
        return resolved_keys



from __future__ import annotations

import fnmatch


def has_permission(required_perm: str, assigned_perms: set[str]):
    if not assigned_perms:
        return False

    if required_perm in assigned_perms or '*' in assigned_perms:
        return True

    glob_patterns = [p for p in assigned_perms if '*' in p]
    for pattern in glob_patterns:
        if fnmatch.fnmatchcase(required_perm, pattern):
            return True
    return False


def get_view_permission(view, method) -> str | None:
    handler = getattr(view, method.lower(), None)
    if handler:
        perm = getattr(handler, '_required_permission', None)
        if perm:
            return perm

    perm_func = getattr(view, 'get_required_permission', None)
    if perm_func:
        return perm_func()

    return getattr(view, 'required_permission', None)



import pytest
from django.test import override_settings
from django.core.management import call_command
from django.core.management.base import CommandError


@override_settings(SAAS_SECRETS_FILE=None)
def test_no_secrets_file():
    with pytest.raises(CommandError, match='SAAS_SECRETS_FILE setting is not set'):
        call_command('secrets', 'list')


@override_settings(SAAS_SECRETS_FILE='.not-exists-secrets')
def test_secrets_list_empty(capsys):
    call_command('secrets', 'list')
    captured = capsys.readouterr()
    assert 'No secrets found.' in captured.out


def test_list_default_secrets(capsys):
    call_command('secrets', 'list')
    captured = capsys.readouterr()
    assert 'turnstile' in captured.out



from rest_framework import serializers
from .email_code import (
    EmailCode,
    EmailCodeRequestSerializer,
    EmailCodeConfirmSerializer,
    NewUserEmailMixin,
)
from ..models import UserEmail

NEW_EMAIL_CODE = 'saas:new_email_code'


class UserEmailSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserEmail
        exclude = ['user']
        read_only_fields = ['id', 'email', 'verified', 'created_at']

    def update(self, instance, validated_data):
        if validated_data.get('primary') and not instance.primary:
            # keep only one primary user email
            UserEmail.objects.filter(user_id=instance.user_id).update(primary=False)
        obj = super().update(instance, validated_data)
        return obj


class AddEmailRequestSerializer(NewUserEmailMixin, EmailCodeRequestSerializer):
    CACHE_PREFIX = NEW_EMAIL_CODE

    def create(self, validated_data) -> EmailCode:
        request = self.context['request']
        code = self.save_auth_code(request.user.pk)
        email = validated_data['email']
        return EmailCode(email, code, request.user)


class AddEmailConfirmSerializer(NewUserEmailMixin, EmailCodeConfirmSerializer):
    CACHE_PREFIX = NEW_EMAIL_CODE

    def validate_code(self, value):
        user_id = super().validate_code(value)
        request = self.context['request']
        if not user_id or request.user.pk != user_id:
            self.fail_code()
        return request.user

    def create(self, validated_data):
        user = validated_data['code']
        return UserEmail.objects.create(
            email=validated_data['email'],
            user=user,
            verified=True,
        )



import typing as t
from django.tasks import task
from django.core.mail import EmailMultiAlternatives


@task
def send_email(
    subject: str,
    recipients: t.List[str],
    text_message: str,
    html_message: t.Optional[str] = None,
    from_email: t.Optional[str] = None,
    headers: t.Optional[t.Dict[str, str]] = None,
    reply_to: t.Optional[str] = None,
):
    mail = EmailMultiAlternatives(
        subject,
        body=text_message,
        from_email=from_email,
        to=recipients,
        headers=headers,
        reply_to=reply_to,
    )
    if html_message:
        mail.attach_alternative(html_message, 'text/html')

    return mail.send()



from typing import List, Any
from drf_spectacular.openapi import AutoSchema as _AutoSchema
from saas_base.registry import perm_registry
from saas_base.registry.checker import get_view_permission


__all__ = ['AutoSchema']


class AutoSchema(_AutoSchema):
    def get_security(self):
        required_perm = get_view_permission(self.view, self.method)
        if not required_perm:
            return []

        required_scopes = perm_registry.get_scopes_for_permission(required_perm)

        if required_scopes:
            return [
                {'oauth2': required_scopes},
            ]

        return []

    def get_filter_backends(self) -> List[Any]:
        return getattr(self.view, 'filter_backends', [])

    def get_description(self) -> str:
        description = super().get_description()
        required_perm = get_view_permission(self.view, self.method)
        if required_perm:
            description = f'**Permissions**: {required_perm}\n\n{description}'
        return description



from django.dispatch import Signal

before_create_tenant = Signal()
before_update_tenant = Signal()
after_signup_user = Signal()
after_login_user = Signal()
member_invited = Signal()
mail_queued = Signal()


__all__ = [
    'before_create_tenant',
    'before_update_tenant',
    'after_signup_user',
    'after_login_user',
    'member_invited',
    'mail_queued',
]



from django.dispatch import receiver
from saas_base.signals import mail_queued
from saas_base.utils.mail_templates import render_mail_messages
from saas_base.tasks.send_mails import send_email


@receiver(mail_queued)
def send_mail_async(sender, template_id, subject, recipients, context, **kwargs):
    text_message, html_message = render_mail_messages(template_id, context)
    send_email.enqueue(
        subject=str(subject),
        recipients=recipients,
        text_message=text_message,
        html_message=html_message,
    )


__import__('saas_base.registry.default_roles')
__import__('saas_base.registry.default_scopes')



from saas_base.models import Invitation, Member
from tests.client import FixturesTestCase


class TestUserInvitationsAPI(FixturesTestCase):
    user_id = FixturesTestCase.GUEST_USER_ID

    def setUp(self):
        super().setUp()
        self.invitation = Invitation.objects.create(
            tenant=self.tenant,
            inviter_id=self.OWNER_USER_ID,
            email=self.user.email,
            role='MEMBER',
            status=Invitation.InviteStatus.SENT,
        )

    def test_list_invitations(self):
        self.force_login()
        url = '/m/user/invitations/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], str(self.invitation.id))

    def test_accept_invitation(self):
        self.force_login()
        url = f'/m/user/invitations/{self.invitation.id}/'
        resp = self.client.patch(url, {'status': 'accepted'})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['status'], 'accepted')

        self.invitation.refresh_from_db()
        self.assertEqual(self.invitation.status, Invitation.InviteStatus.ACCEPTED)
        self.assertIsNotNone(self.invitation.accepted_at)

        member = Member.objects.get(tenant=self.tenant, user=self.user)
        self.assertEqual(member.role, 'MEMBER')

    def test_accept_invitation_not_yours(self):
        self.invitation.email = 'other@example.com'
        self.invitation.save()

        self.force_login()
        url = f'/m/user/invitations/{self.invitation.id}/'
        resp = self.client.patch(url, {'status': 'accepted'})
        self.assertEqual(resp.status_code, 403)

    def test_accept_already_member(self):
        Member.objects.create(tenant=self.tenant, user=self.user, role='MEMBER')

        self.force_login()
        url = f'/m/user/invitations/{self.invitation.id}/'
        resp = self.client.patch(url, {'status': 'accepted'})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['status'], 'accepted')



from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django.contrib.auth import login, logout
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle
from rest_framework.exceptions import NotFound
from saas_base.models import Invitation
from saas_base.drf.views import Endpoint
from saas_base.settings import saas_settings
from saas_base.rules import check_rules
from saas_base.serializers.auth import (
    EmailCode,
    SignupRequestCodeSerializer,
    SignupCreateUserSerializer,
    SignupConfirmCodeSerializer,
    SignupConfirmPasswordSerializer,
    SignupWithInvitationSerializer,
)
from saas_base.serializers.invitation import InvitationInfoSerializer
from saas_base.serializers.password import PasswordLoginSerializer
from saas_base.signals import after_signup_user, after_login_user, mail_queued


__all__ = [
    'SignupRequestEndpoint',
    'SignupConfirmEndpoint',
    'SignupWithInvitationEndpoint',
    'PasswordLogInEndpoint',
    'LogoutEndpoint',
    'InvitationEndpoint',
]


class SignupRequestEndpoint(Endpoint):
    email_template_id = 'signup_code'
    email_subject = _('Signup Request')
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]

    def get_serializer_class(self):
        if saas_settings.SIGNUP_REQUEST_CREATE_USER:
            return SignupCreateUserSerializer
        return SignupRequestCodeSerializer

    def post(self, request: Request):
        """Send a sign-up code to user's email address."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        # check bad request rules
        check_rules(saas_settings.SIGNUP_SECURITY_RULES, request)

        obj: EmailCode = serializer.save()
        mail_queued.send(
            sender=self.__class__,
            template_id=self.email_template_id,
            subject=str(self.email_subject),
            recipients=[obj.recipient()],
            context={'code': obj.code},
            request=request,
        )
        return Response(status=204)


class _BaseSignupConfirmEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]

    def post(self, request: Request, *args, **kwargs):
        """Register a new user and login."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.save()

        # auto login after signup
        login(request._request, user, backend='saas_base.auth.backends.ModelBackend')
        after_signup_user.send(
            self.__class__,
            user=user,
            request=request,
            strategy='password',
        )
        return Response({'next': settings.LOGIN_REDIRECT_URL})


class SignupConfirmEndpoint(_BaseSignupConfirmEndpoint):
    def get_serializer_class(self):
        if saas_settings.SIGNUP_REQUEST_CREATE_USER:
            return SignupConfirmCodeSerializer
        return SignupConfirmPasswordSerializer


class SignupWithInvitationEndpoint(_BaseSignupConfirmEndpoint):
    serializer_class = SignupWithInvitationSerializer
    queryset = Invitation.objects.all()

    def get_serializer_context(self):
        obj: Invitation = self.get_object()
        # only allow signup with "SENT" status
        if obj.status == Invitation.InviteStatus.SENT:
            context = super().get_serializer_context()
            context['invitation'] = obj
            return context
        raise NotFound()


class PasswordLogInEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordLoginSerializer

    def post(self, request: Request):
        """Login a user with the given username and password."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        check_rules(saas_settings.LOGIN_SECURITY_RULES, request)

        user = serializer.save()
        login(request._request, user, backend='saas_base.auth.backends.ModelBackend')

        after_login_user.send(
            self.__class__,
            user=user,
            request=request,
            strategy='password',
        )
        return Response({'next': settings.LOGIN_REDIRECT_URL})


class LogoutEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []

    def post(self, request: Request):
        """Clear the user session and log the user out."""
        logout(request._request)
        return Response({'next': settings.LOGIN_URL})


class InvitationEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    serializer_class = InvitationInfoSerializer
    queryset = Invitation.objects.all()

    def get(self, request: Request, *args, **kwargs):
        """Retrieve a pending membership invitation."""
        obj: Invitation = self.get_object()
        if obj.status == Invitation.InviteStatus.ACCEPTED:
            raise NotFound()
        serializer = self.get_serializer(obj)
        return Response(serializer.data)



from django.db.models.signals import post_save
from django.dispatch import receiver
from saas_base.models import get_tenant_model
from saas_base.models import Group, Member, UserEmail

Tenant = get_tenant_model()


@receiver(post_save, sender=Tenant)
def ensure_owner_is_member(sender, instance: Tenant, *args, **kwargs):
    Member.objects.get_or_create(
        tenant=instance,
        user_id=instance.owner_id,
        defaults={'role': 'OWNER'},
    )


def setup():
    print('Setup Group, Member, UserEmail cache invalidation')
    Group.objects.setup_related_cache_invalidation()
    Member.objects.setup_related_cache_invalidation()
    UserEmail.objects.setup_related_cache_invalidation()

    print('Setup default permissions')
    __import__('saas_base.registry.default_perms')



from io import StringIO
from django.core.management import call_command
from saas_base.models import Member
from tests.client import FixturesTestCase


class TestSyncTenantOwners(FixturesTestCase):
    def test_sync_tenant_owners_creates_missing_member(self):
        Member.objects.all().delete()
        tenant = self.tenant
        self.assertFalse(Member.objects.filter(tenant=tenant, user=tenant.owner).exists())

        call_command('sync_tenant_owners')

        member = Member.objects.get(tenant=tenant, user=tenant.owner)
        self.assertEqual(member.role, 'OWNER')

    def test_sync_tenant_owners_updates_wrong_role(self):
        tenant = self.tenant
        member = Member.objects.get(tenant=tenant, user=tenant.owner)
        member.role = 'MEMBER'
        member.save()

        call_command('sync_tenant_owners')

        member.refresh_from_db()
        self.assertEqual(member.role, 'OWNER')

    def test_sync_tenant_owners_no_changes_needed(self):
        tenant = self.tenant
        member = Member.objects.get(tenant=tenant, user=tenant.owner)
        self.assertEqual(member.role, 'OWNER')

        # Capture output to verify counts
        out = StringIO()
        call_command('sync_tenant_owners', stdout=out)

        output = out.getvalue()
        self.assertIn('Created 0 missing owner members', output)
        self.assertIn('Updated 0 owner roles', output)



from django.utils.translation import gettext as _
from rest_framework import serializers
from saas_base.drf.serializers import ModelSerializer, RelatedSerializerField
from saas_base.models import Member
from saas_base.registry import perm_registry
from saas_base.registry.checker import has_permission
from .group import GroupSerializer
from .user import UserSerializer


class MemberSerializer(ModelSerializer):
    user = UserSerializer(required=False, read_only=True)
    groups = serializers.SlugRelatedField(many=True, read_only=True, slug_field='name')
    permissions = serializers.SerializerMethodField()

    class Meta:
        model = Member
        exclude = ['tenant']
        request_include_fields = ['user', 'groups']

    def get_permissions(self, obj: Member):
        assigned_patterns = obj.get_all_permissions()

        all_registered_perms = perm_registry.get_permission_list()
        if obj.tenant.owner_id == obj.user_id:
            return [p.key for p in all_registered_perms]

        resolved_keys = [p.key for p in all_registered_perms if has_permission(p.key, assigned_patterns)]
        return resolved_keys


class MemberUpdateSerializer(ModelSerializer):
    groups = RelatedSerializerField(GroupSerializer, many=True)

    class Meta:
        model = Member
        fields = ['role', 'groups', 'permissions']

    def validate_role(self, role: str):
        if role not in perm_registry.get_role_codes():
            raise serializers.ValidationError(_('Invalid role.'))
        return role

    def validate_permissions(self, perms: list[str]):
        registered_perms = perm_registry.get_permission_keys()
        for key in perms:
            if not has_permission(key, registered_perms):
                raise serializers.ValidationError(f'Permission {key} is not registered.')
        return perms



from .base import Rule, check_rules
from .email_address import (
    load_default_blocked_domains,
    BlockedEmailDomains,
    AvoidTooManyDots,
    AvoidUsingPlus,
)
from .turnstile import Turnstile


__all__ = [
    'Rule',
    'check_rules',
    'BlockedEmailDomains',
    'AvoidTooManyDots',
    'AvoidUsingPlus',
    'Turnstile',
    'load_default_blocked_domains',
]



from django.utils.translation import gettext_lazy as _
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.mixins import (
    ListModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
)
from saas_base.drf.views import TenantEndpoint
from saas_base.drf.decorators import resource_permission
from saas_base.drf.filters import TenantIdFilter, IncludeFilter
from saas_base.serializers.invitation import InvitationSerializer
from saas_base.models import Invitation
from saas_base.settings import saas_settings
from saas_base.signals import member_invited, mail_queued

__all__ = [
    'InvitationListEndpoint',
    'InvitationItemEndpoint',
]


class InvitationListEndpoint(ListModelMixin, TenantEndpoint):
    email_template_id = 'invite_member'
    email_subject = _("You've Been Invited to Join %s")

    serializer_class = InvitationSerializer
    filter_backends = [TenantIdFilter, IncludeFilter]
    queryset = Invitation.objects.all()
    pagination_class = None

    include_select_related_fields = ['role']
    include_prefetch_related_fields = ['groups']

    def get_queryset(self):
        return self.queryset.filter(status__lt=Invitation.InviteStatus.ACCEPTED)

    def get_email_subject(self):
        return self.email_subject % str(self.request.tenant)

    @resource_permission('iam.invitation.view')
    def get(self, request: Request, *args, **kwargs):
        """List all members in the tenant."""
        return self.list(request, *args, **kwargs)

    @resource_permission('iam.invitation.manage')
    def post(self, request: Request, *args, **kwargs):
        """Invite a member to join the tenant."""
        tenant_id = self.get_tenant_id()
        context = self.get_serializer_context()
        serializer = InvitationSerializer(data=request.data, context=context)
        serializer.is_valid(raise_exception=True)
        invitation = serializer.save(tenant_id=tenant_id, inviter=request.user)

        member_invited.send(self.__class__, member=invitation, request=request)

        recipient = invitation.email
        invite_link = saas_settings.INVITATION_LINK % str(invitation.id)
        if not invite_link.startswith('http'):
            invite_link = request.build_absolute_uri(invite_link)

        mail_queued.send(
            sender=self.__class__,
            template_id=self.email_template_id,
            subject=str(self.get_email_subject()),
            recipients=[recipient],
            context={
                'inviter': request.user,
                'member': invitation,
                'tenant': request.tenant,
                'invite_link': invite_link,
            },
            request=request,
        )
        return Response(serializer.data, status=201)


class InvitationItemEndpoint(UpdateModelMixin, DestroyModelMixin, TenantEndpoint):
    serializer_class = InvitationSerializer
    queryset = Invitation.objects.all()

    @resource_permission('iam.invitation.manage')
    def patch(self, request: Request, *args, **kwargs):
        """Update an invitation's role and groups."""
        return self.partial_update(request, *args, **kwargs)

    @resource_permission('iam.invitation.manage')
    def delete(self, request: Request, *args, **kwargs):
        """Remove an invitation from the tenant."""
        return self.destroy(request, *args, **kwargs)



import typing as t
import css_inline
from django.template import loader
from saas_base.settings import saas_settings


def render_mail_messages(template_id: str, context: t.Dict[str, t.Any]) -> t.Tuple[str, str]:
    context.setdefault('site', saas_settings.SITE)
    text: str = loader.render_to_string(f'saas_emails/{template_id}.text', context)
    html: str = loader.render_to_string(f'saas_emails/{template_id}.html', context)
    return text, css_inline.inline(html)



from dataclasses import asdict
from rest_framework.views import APIView
from rest_framework.request import Request
from rest_framework.response import Response
from saas_base.registry import perm_registry


__all__ = ['RoleListEndpoint']


class RoleListEndpoint(APIView):
    permission_classes = []
    pagination_class = None

    def get(self, request: Request, *args, **kwargs):
        """Show all supported Roles."""
        data = [asdict(p) for p in perm_registry.get_role_list()]
        return Response(data)



from rest_framework import serializers
from drf_spectacular.utils import extend_schema
from drf_spectacular.extensions import OpenApiViewExtension
from saas_base.serializers._spectacular import PermissionSerializer, RoleSerializer


class AuthResponseSerializer(serializers.Serializer):
    next = serializers.CharField()


class FixedPermissionSerializer(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.permissions.PermissionListEndpoint'

    def view_replacement(self):
        class PermissionListEndpoint(self.target_class):
            @extend_schema(summary='All Permissions', responses={200: PermissionSerializer})
            def get(self, *args, **kwargs):
                pass

        return PermissionListEndpoint


class FixedRoleSerializer(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.roles.RoleListEndpoint'

    def view_replacement(self):
        class RoleListEndpoint(self.target_class):
            @extend_schema(summary='All Roles', responses={200: RoleSerializer})
            def get(self, *args, **kwargs):
                pass

        return RoleListEndpoint


class FixedPasswordLogInEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.PasswordLogInEndpoint'

    def view_replacement(self):
        class PasswordLogInEndpoint(self.target_class):
            @extend_schema(summary='Log In', responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return PasswordLogInEndpoint


class FixedLogoutEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.LogoutEndpoint'

    def view_replacement(self):
        class LogoutEndpoint(self.target_class):
            @extend_schema(summary='Log Out', request=None, responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return LogoutEndpoint


class FixedSignupConfirmEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.SignupConfirmEndpoint'

    def view_replacement(self):
        class SignupConfirmEndpoint(self.target_class):
            @extend_schema(summary='Sign Up', responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return SignupConfirmEndpoint


class FixedSignupRequestEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.SignupRequestEndpoint'

    def view_replacement(self):
        class SignupRequestEndpoint(self.target_class):
            @extend_schema(summary='Request to Sign-up', responses={204: None})
            def post(self, *args, **kwargs):
                pass

        return SignupRequestEndpoint


class FixedPasswordResetEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.password.PasswordResetEndpoint'

    def view_replacement(self):
        class PasswordResetEndpoint(self.target_class):
            @extend_schema(summary='Password Reset', responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return PasswordResetEndpoint


class FixedPasswordForgotEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.password.PasswordForgotEndpoint'

    def view_replacement(self):
        class PasswordForgotEndpoint(self.target_class):
            @extend_schema(summary='Password Forgot', responses={204: None})
            def post(self, *args, **kwargs):
                pass

        return PasswordForgotEndpoint


class FixedUserPasswordEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.user.UserPasswordEndpoint'

    def view_replacement(self):
        class UserPasswordEndpoint(self.target_class):
            @extend_schema(summary='Update Password')
            def post(self, *args, **kwargs):
                pass

        return UserPasswordEndpoint


class FixedUserEmailItemEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.user_emails.UserEmailItemEndpoint'

    def view_replacement(self):
        class UserEmailItemEndpoint(self.target_class):
            @extend_schema(summary='View Email Details')
            def get(self, *args, **kwargs):
                pass

        return UserEmailItemEndpoint



from saas_base.models import get_tenant_model, Member
from tests.client import FixturesTestCase


Tenant = get_tenant_model()


class TestUserTenantsAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_list_tenants(self):
        self.force_login()
        url = '/m/user/tenants/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        tenants = resp.json()
        self.assertEqual(len(tenants), 2)
        self.assertIn('role', tenants[0])
        self.assertNotIn('member_count', tenants[0])
        self.assertNotIn('groups', tenants[0])

        url = '/m/user/tenants/?include=groups,member_count'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        tenants = resp.json()
        self.assertEqual(len(tenants), 2)
        self.assertIn('role', tenants[0])
        self.assertIn('groups', tenants[0])
        self.assertEqual(tenants[0]['member_count'], 1)

        self.force_login(self.GUEST_USER_ID)
        resp = self.client.get(url)
        self.assertEqual(len(resp.json()), 0)

    def test_leave_tenant(self):
        self.force_login()
        resp = self.client.delete(f'/m/user/tenants/{self.tenant_id}/')
        self.assertEqual(resp.status_code, 403)

        self.force_login(self.GUEST_USER_ID)
        Member.objects.create(user_id=self.GUEST_USER_ID, tenant_id=self.tenant_id)
        resp = self.client.delete(f'/m/user/tenants/{self.tenant_id}/')
        self.assertEqual(resp.status_code, 204)



from tests.client import FixturesTestCase


class TestRolesAPI(FixturesTestCase):
    def test_list_all_roles(self):
        resp = self.client.get('/m/roles/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIsInstance(data, list)
        role_names = [d['name'] for d in data]
        self.assertIn('Owner', role_names)



from rest_framework import serializers
from saas_base.registry import perm_registry
from saas_base.registry.checker import has_permission
from saas_base.models import Group


class GroupSerializer(serializers.ModelSerializer):
    class Meta:
        model = Group
        exclude = ['tenant']

    def validate_permissions(self, perms: list[str]):
        registered_perms = perm_registry.get_permission_keys()
        for key in perms:
            if not has_permission(key, registered_perms):
                raise serializers.ValidationError(f'Permission {key} is not registered.')
        return perms



from rest_framework.request import Request
from rest_framework.response import Response
from saas_base.drf.views import AuthenticatedEndpoint
from saas_base.drf.decorators import resource_permission
from ..serializers.user import (
    UserSerializer,
    UserPasswordSerializer,
)

__all__ = [
    'UserEndpoint',
    'UserPasswordEndpoint',
]


class UserEndpoint(AuthenticatedEndpoint):
    serializer_class = UserSerializer

    @resource_permission('user.profile.view')
    def get(self, request: Request):
        """Retrieve current user information."""
        serializer: UserSerializer = self.get_serializer(request.user)
        return Response(serializer.data)

    @resource_permission('user.profile.update')
    def patch(self, request, *args, **kwargs):
        """Update current user information."""
        serializer = self.get_serializer(request.user, data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)


class UserPasswordEndpoint(AuthenticatedEndpoint):
    serializer_class = UserPasswordSerializer

    # this is an non-exist permission, so that API token can not
    # access this endpoint
    required_permission = 'user.password.update'

    def post(self, request: Request, *args, **kwargs):
        """Update current user's password"""
        serializer: UserPasswordSerializer = self.get_serializer(request.user, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(status=204)



# Generated by Django 6.0.1 on 2026-01-13 06:48

import django.db.models.deletion
import django.utils.timezone
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('saas_base', '0005_userprofile_alter_useremail_table'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveField(
            model_name='member',
            name='permissions',
        ),
        migrations.RemoveField(
            model_name='group',
            name='permissions',
        ),
        migrations.RemoveField(
            model_name='role',
            name='permissions',
        ),
        migrations.AlterField(
            model_name='member',
            name='role',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.RemoveField(
            model_name='member',
            name='email',
        ),
        migrations.RemoveField(
            model_name='member',
            name='inviter',
        ),
        migrations.RemoveField(
            model_name='member',
            name='name',
        ),
        migrations.RemoveField(
            model_name='member',
            name='status',
        ),
        migrations.AlterField(
            model_name='member',
            name='groups',
            field=models.ManyToManyField(blank=True, to='saas_base.group'),
        ),
        migrations.AlterField(
            model_name='member',
            name='user',
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterField(
            model_name='tenant',
            name='owner',
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.CreateModel(
            name='Invitation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('email', models.EmailField(max_length=254)),
                ('role', models.CharField(blank=True, max_length=50, null=True)),
                (
                    'status',
                    models.IntegerField(
                        choices=[(1, 'pending'), (2, 'sent'), (3, 'accepted'), (4, 'expired')], default=1
                    ),
                ),
                ('accepted_at', models.DateTimeField(blank=True, null=True)),
                ('created_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ('groups', models.ManyToManyField(blank=True, to='saas_base.group')),
                (
                    'inviter',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='sent_invitations',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'tenant',
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.SAAS_TENANT_MODEL),
                ),
            ],
            options={
                'verbose_name': 'invitation',
                'verbose_name_plural': 'invitations',
                'db_table': 'saas_member_invitation',
                'ordering': ['-created_at'],
                'unique_together': {('tenant', 'email')},
            },
        ),
        migrations.AddField(
            model_name='member',
            name='permissions',
            field=models.JSONField(default=list),
        ),
        migrations.AddField(
            model_name='group',
            name='permissions',
            field=models.JSONField(default=list),
        ),
        migrations.DeleteModel(
            name='Permission',
        ),
        migrations.DeleteModel(
            name='Role',
        ),
    ]



from django.utils import timezone
from saas_base.test import SaasTestCase
from saas_base.models import Member, Group, get_tenant_model


Tenant = get_tenant_model()


class TestUserSecrets(SaasTestCase):
    user_id = SaasTestCase.OWNER_USER_ID

    def test_update_foreign_key(self):
        member = Member.objects.get_from_cache_by_natural_key(self.tenant_id, self.user_id)
        self.assertEqual(member.user.username, 'demo-1')

        user = self.get_user()
        user.username = 'changed-1'
        user.save()

        member = Member.objects.get_from_cache_by_natural_key(self.tenant_id, self.user_id)
        self.assertEqual(member.user.username, 'changed-1')

    def test_update_m2m(self):
        member = Member.objects.get_from_cache_by_natural_key(self.tenant_id, self.user_id)
        perms = member.get_all_permissions()
        self.assertEqual(perms, {'*'})

        group = Group.objects.create(tenant_id=self.tenant_id, name='Admin', permissions=['org.info.view'])
        member.groups.add(group)

        member = Member.objects.get_from_cache_by_natural_key(self.tenant_id, self.user_id)
        perms = member.get_all_permissions()
        self.assertEqual(perms, {'*', 'org.info.view'})

    def test_update_field(self):
        tenant = Tenant.objects.get_from_cache_by_pk(self.tenant_id)
        tenant.name = 'Changed'
        self.assertEqual(tenant.name, 'Changed')

        tenant = Tenant.objects.get_from_cache_by_pk(self.tenant_id)
        self.assertIsNone(tenant.expires_at)
        self.assertNotEqual(tenant.name, 'Changed')

        tenant.name = 'Changed'
        tenant.expires_at = timezone.now()
        tenant.save()

        tenant = Tenant.objects.get_from_cache_by_pk(self.tenant_id)
        self.assertEqual(tenant.name, 'Changed')
        self.assertIsNotNone(tenant.expires_at)



from rest_framework import serializers


class PermissionSerializer(serializers.Serializer):
    key = serializers.CharField()
    label = serializers.CharField()
    description = serializers.CharField()
    module = serializers.CharField()


class RoleSerializer(serializers.Serializer):
    code = serializers.CharField()
    name = serializers.CharField()
    description = serializers.CharField()
    permissions = serializers.ListField(child=serializers.CharField())



from django.urls import path
from ..endpoints.user_invitations import (
    UserInvitationListEndpoint,
    UserInvitationAcceptEndpoint,
)

urlpatterns = [
    path('', UserInvitationListEndpoint.as_view()),
    path('<pk>/', UserInvitationAcceptEndpoint.as_view()),
]



from django.test import TestCase
from django.core.management import call_command
from django.contrib.auth import get_user_model
from rest_framework.test import APIClient
from saas_base.models import get_tenant_model


class TestMembersAPI(TestCase):
    fixtures = [
        'test_data.yaml',
    ]
    client: APIClient
    client_class = APIClient

    tenant_id = 1
    user_id = 3

    def get_user(self):
        return get_user_model().objects.get(pk=self.user_id)

    @property
    def tenant(self):
        return get_tenant_model().objects.get(pk=self.tenant_id)

    def test_list_users_via_owner(self):
        call_command('sync_tenant_owners')
        self.client.force_login(self.get_user())
        resp = self.client.get(f'/tenant/{self.tenant_id}/members/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['count'], 1)



import typing as t
from collections import OrderedDict, defaultdict

from django.db.models import QuerySet
from django.core.exceptions import ObjectDoesNotExist
from rest_framework.validators import ValidationError
from rest_framework.fields import Field, ChoiceField as _ChoiceField
from rest_framework.serializers import Serializer, ModelSerializer as _ModelSerializer


__all__ = [
    'ChoiceField',
    'RelatedSerializerField',
    'ModelSerializer',
    'FlattenModelSerializer',
]


class ChoiceField(_ChoiceField):
    """Rewrite this fields to support (int, str) choices."""

    def __init__(self, choices, **kwargs):
        super().__init__(choices, **kwargs)

        self.int_str_choices: t.Dict[int, str] = {}
        self.str_int_choices: t.Dict[str, int] = {}
        if choices:
            self._set_int_str_choices(choices)

    def _set_int_str_choices(self, choices: t.List[t.Tuple[int, str]]):
        int_str_choices: t.Dict[int, str] = {}
        str_int_choices: t.Dict[str, int] = {}
        for choice in choices:
            if len(choice) == 2 and isinstance(choice[0], int):
                int_str_choices[choice[0]] = choice[1]
                str_int_choices[choice[1]] = choice[0]
            else:
                return
        self.int_str_choices = int_str_choices
        self.str_int_choices = str_int_choices

    def to_internal_value(self, data):
        if data == '' and self.allow_blank:
            return ''

        if self.str_int_choices:
            try:
                return self.str_int_choices[str(data)]
            except KeyError:
                self.fail('invalid_choice', input=data)

        try:
            return self.choice_strings_to_values[str(data)]
        except KeyError:
            self.fail('invalid_choice', input=data)

    def to_representation(self, value):
        if value in ('', None):
            return value

        if self.int_str_choices:
            return self.int_str_choices[value]

        return self.choice_strings_to_values.get(str(value), value)


class RelatedSerializerField(Field):
    def __init__(self, serializer_cls, **kwargs):
        self.serializer_cls = serializer_cls
        self.many = kwargs.pop('many', False)
        super().__init__(**kwargs)

    def to_representation(self, value: QuerySet):
        return self.serializer_cls(value, many=self.many).data

    def to_internal_value(self, data):
        model = self.serializer_cls.Meta.model
        if self.many and not isinstance(data, list):
            raise ValidationError(f'Expected a list of {model.__name__} IDs.')

        if self.many:
            return model.objects.filter(pk__in=data)
        try:
            return model.objects.get(pk=data)
        except model.DoesNotExist:
            raise ValidationError(f'Invalid {model.__name__} ID.')


class ModelSerializer(_ModelSerializer):
    serializer_choice_field = ChoiceField

    @property
    def _readable_fields(self):
        request_include_fields = getattr(self.Meta, 'request_include_fields', [])
        if not request_include_fields:
            for field in self.fields.values():
                if not field.write_only:
                    yield field
        else:
            include_terms = self.context.get('include_fields', [])
            if not include_terms:
                request = self.context.get('request')
                if request:
                    include_terms = getattr(request, 'include_terms', [])

            for field in self.fields.values():
                if field.field_name in request_include_fields and field.field_name not in include_terms:
                    continue

                if not field.write_only:
                    yield field


class FlattenModelSerializer(ModelSerializer):
    @property
    def _flatten_fields(self) -> t.List[str]:
        return getattr(self.Meta, 'flatten_fields', [])

    def get_fields(self):
        fields = super().get_fields()
        if not self._flatten_fields:
            return fields

        new_fields = OrderedDict()
        for root_name, field in fields.items():
            if root_name in self._flatten_fields and isinstance(field, Serializer):
                field.field_name = root_name
                nested_fields = field.get_fields()
                for nested_name, nested_field in nested_fields.items():
                    nested_field._flatten_parent = field
                    if nested_field.source:
                        nested_field.source = f'{root_name}.{nested_field.source}'
                    else:
                        nested_field.source = f'{root_name}.{nested_name}'

                    # Handle naming collisions: if field exists in root, prefix it
                    if nested_name in fields:
                        new_fields[f'{root_name}_{nested_name}'] = nested_field
                    else:
                        new_fields[nested_name] = nested_field
            else:
                new_fields[root_name] = field
        return new_fields

    def to_internal_value(self, data):
        if not self._flatten_fields:
            return super().to_internal_value(data)

        internal_data = data.copy()
        nested_data = defaultdict(dict)
        nested_serializers = {}
        for field_name, field in self.fields.items():
            if hasattr(field, '_flatten_parent'):
                parent_field = getattr(field, '_flatten_parent')
                root_name = parent_field.field_name
                nested_serializers[root_name] = parent_field
                nested_key = f'{root_name}_{field_name}'
                if nested_key in internal_data:
                    nested_data[root_name][field_name] = internal_data.pop(nested_key)
                elif field_name in internal_data:
                    nested_data[root_name][field_name] = internal_data.pop(field_name)

        result = super().to_internal_value(internal_data)

        # Validate nested fields
        for root_name, nested_data in nested_data.items():
            serializer = nested_serializers[root_name]
            validated_data = serializer.to_internal_value(nested_data)
            result[root_name] = validated_data
        return result

    @staticmethod
    def _get_related_field_name(instance, model):
        for field in model._meta.get_fields():
            # Check if this field is a relation pointing to our parent model
            if field.is_relation and field.related_model == instance.__class__:
                return field.name

        # Fallback or Raise error if no relation is found
        raise AttributeError(f'No relation from {model.__name__} back to {instance.__class__.__name__}.')

    def _update_nested_fields(self, instance, data: t.Dict[str, t.Any]):
        nested_serializers = {}
        for field in self.fields.values():
            if hasattr(field, '_flatten_parent'):
                parent_field = getattr(field, '_flatten_parent')
                nested_serializers[parent_field.field_name] = parent_field

        for name, value in data.items():
            serializer_field = nested_serializers[name]
            try:
                related_instance = getattr(instance, name)
                serializer_field.update(related_instance, value)
            except (ObjectDoesNotExist, AttributeError):
                nested_model = serializer_field.Meta.model
                instance_name = self._get_related_field_name(instance, nested_model)
                value[instance_name] = instance
                serializer_field.create(value)

    def create(self, validated_data):
        nested_data = {f: validated_data.pop(f) for f in self._flatten_fields if f in validated_data}
        obj = super().create(validated_data)
        self._update_nested_fields(obj, nested_data)
        return obj

    def update(self, instance, validated_data):
        nested_data = {f: validated_data.pop(f) for f in self._flatten_fields if f in validated_data}
        obj = super().update(instance, validated_data)
        self._update_nested_fields(obj, nested_data)
        return obj



from unittest.mock import Mock
from django.utils import timezone
from django.test import RequestFactory
from rest_framework.request import Request
from rest_framework.exceptions import PermissionDenied
from saas_base.endpoints.tenant import TenantItemEndpoint
from saas_base.drf.permissions import (
    IsTenantOwner,
    IsTenantOwnerOrReadOnly,
    IsTenantActive,
    IsTenantActiveOrReadOnly,
    HasResourcePermission,
)
from saas_base.models import Member, get_tenant_model
from tests.demo_app.models import TenantSecret
from tests.client import FixturesTestCase

Tenant = get_tenant_model()


class TestIsTenantOwner(FixturesTestCase):
    def setUp(self):
        super().setUp()
        self.permission = IsTenantOwner()
        self.view = Mock()
        factory = RequestFactory()
        self.request = Request(factory.get('/'))
        self.request.tenant_id = self.tenant_id

    def test_has_permission_owner(self):
        self.request.user = self.force_login(self.OWNER_USER_ID)
        self.request.tenant_id = self.tenant_id
        self.request.member = Member.objects.get(tenant=self.tenant, user_id=self.OWNER_USER_ID)
        self.assertTrue(self.permission.has_permission(self.request, self.view))

    def test_has_permission_member(self):
        self.request.user = self.force_login(self.MEMBER_USER_ID)
        self.assertFalse(self.permission.has_permission(self.request, self.view))

        # readonly passes
        permission = IsTenantOwnerOrReadOnly()
        self.assertTrue(permission.has_permission(self.request, self.view))

    def test_has_permission_member_with_owner_role(self):
        Member.objects.create(tenant=self.tenant, user_id=self.MEMBER_USER_ID, role='OWNER')
        user = self.force_login(self.MEMBER_USER_ID)
        self.request.user = user
        # Mock member attribute on request
        self.request.member = Member.objects.get(tenant=self.tenant, user=user)
        self.assertTrue(self.permission.has_permission(self.request, self.view))

    def test_has_object_permission(self):
        user = self.force_login(self.OWNER_USER_ID)
        self.request.user = user
        self.request.tenant_id = None
        self.view.object_tenant_id_field = 'tenant_id'
        obj = TenantSecret.objects.create(tenant_id=self.tenant_id)
        self.assertTrue(self.permission.has_object_permission(self.request, self.view, obj))


class TestIsTenantActive(FixturesTestCase):
    def setUp(self):
        super().setUp()
        self.permission = IsTenantActive()
        self.view = Mock()
        factory = RequestFactory()
        self.request = Request(factory.get('/'))
        self.request.tenant_id = self.tenant_id

    def test_active_tenant(self):
        user = self.force_login(self.MEMBER_USER_ID)
        self.request.user = user
        self.request.member = Member.objects.create(tenant=self.tenant, user_id=self.MEMBER_USER_ID)
        self.assertTrue(self.permission.has_permission(self.request, self.view))

    def test_expired_tenant(self):
        tenant = self.get_tenant(self.tenant_id)
        tenant.expires_at = timezone.now() - timezone.timedelta(days=1)
        tenant.save()
        user = self.force_login(self.MEMBER_USER_ID)
        self.request.user = user

        with self.assertRaises(PermissionDenied):
            self.permission.has_permission(self.request, self.view)

        # readonly passes
        permission = IsTenantActiveOrReadOnly()
        self.assertTrue(permission.has_permission(self.request, self.view))


class TestHasResourcePermission(FixturesTestCase):
    def setUp(self):
        super().setUp()
        self.permission = HasResourcePermission()
        self.view = TenantItemEndpoint.as_view()
        self.view.required_permission = 'org.info.view'
        factory = RequestFactory()
        self.request = Request(factory.get('/'))
        self.request._auth = None
        self.request.tenant_id = self.tenant_id

    def test_has_permission_owner(self):
        user = self.force_login(self.OWNER_USER_ID)
        self.request.user = user
        self.assertTrue(self.permission.has_permission(self.request, self.view))

    def test_has_permission_member_with_perm(self):
        user = self.force_login(self.MEMBER_USER_ID)
        member = Member.objects.create(tenant=self.tenant, user=user, role='MEMBER')
        self.request.user = user
        self.assertTrue(self.permission.has_permission(self.request, self.view))

        # Test without permission
        member.role = None
        member.save()
        self.assertFalse(self.permission.has_permission(self.request, self.view))

    def test_has_permission_with_invalid_token_scope(self):
        user = self.force_login(self.MEMBER_USER_ID)
        self.request.user = user

        class Token1:
            scope = 'profile'

        self.request._auth = Token1
        self.assertFalse(self.permission.has_permission(self.request, self.view))

        class Token2:
            scopes = ['profile']

        self.request._auth = Token2
        self.assertFalse(self.permission.has_permission(self.request, self.view))

        class Token3:
            def get_scopes(self):
                return ['profile']

        self.request._auth = Token3()
        self.assertFalse(self.permission.has_permission(self.request, self.view))

    def test_has_permission_with_valid_token_scope(self):
        user = self.force_login(self.MEMBER_USER_ID)
        self.request.user = user
        Member.objects.create(tenant=self.tenant, user=user, role='MEMBER')

        class Token1:
            scope = 'org:read'

        self.request._auth = Token1
        self.assertTrue(self.permission.has_permission(self.request, self.view))

        class Token2:
            scopes = ['org:read']

        self.request._auth = Token2
        self.assertTrue(self.permission.has_permission(self.request, self.view))



from tests.client import FixturesTestCase


class TestTenantsAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_list_tenants(self):
        self.force_login()
        resp = self.client.get('/m/tenants/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIsInstance(data, list)

    def test_create_tenant(self):
        self.force_login()
        data = {'name': 'Demo', 'slug': 'demo'}
        resp = self.client.post('/m/tenants/', data=data)
        self.assertEqual(resp.status_code, 201)
        data = resp.json()
        self.assertEqual(data['name'], 'Demo')

    def test_fetch_tenant(self):
        self.force_login()
        resp = self.client.get('/m/tenants/1/')
        self.assertEqual(resp.status_code, 200)

    def test_update_tenant(self):
        self.force_login()
        data = {'name': 'Demo 2'}
        resp = self.client.patch('/m/tenants/1/', data=data)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['name'], 'Demo 2')

    def test_cannot_update_readonly_fields(self):
        self.force_login()
        data = {'slug': 'demo-2', 'region': 'uk'}
        resp = self.client.patch('/m/tenants/1/', data=data)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['slug'], 'demo-1')
        self.assertEqual(data['region'], 'us')


class TestTenantsUsingGuestUser(FixturesTestCase):
    user_id = FixturesTestCase.GUEST_USER_ID

    def test_fetch_tenant(self):
        self.force_login()
        resp = self.client.get('/m/tenants/1/')
        self.assertEqual(resp.status_code, 403)

    def test_update_tenant(self):
        self.force_login()
        data = {'name': 'Demo 2'}
        resp = self.client.patch('/m/tenants/1/', data=data)
        self.assertEqual(resp.status_code, 403)



# Generated by Django 5.2.6 on 2025-11-20 08:00

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('saas_base', '0002_add_role_and_update_member'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='member',
            name='is_staff',
        ),
    ]
